/*
 * *********************************************************
 * Copyright (c) 2009 - 2010, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 21.07.2010
 * Author(s): Gero
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;

import org.apache.log4j.Logger;

import com.sleepycat.persist.model.Persistent;

import edu.dhbw.mannheim.tigers.sumatra.model.SumatraModel;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.SumatraMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2f;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3f;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BallID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.kalman.WPConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.kalman.oextkal.data.BallMotionResult;


/**
 * Simple data holder describing balls that are recognized and tracked by the
 * {@link edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AWorldPredictor}
 * <p>
 * <i>(Being aware of EJ-SE Items 13, 14 and 55: members are public to reduce noise)</i>
 * </p>
 * 
 * @see ATrackedObject
 * @author Gero
 */
@Persistent
public class TrackedBall extends ATrackedObject
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	private static final Logger	log						= Logger.getLogger(TrackedBall.class
																				.getName());
	
	
	/** mm */
	private Vector3					pos;
	/** m/s */
	private Vector3					vel;
	/** m/s^2 */
	private Vector3					acc;
	
	/** not final for ObjectDB */
	private boolean					onCam;
	
	private static float				p00						= 0;
	private static float				p10						= 0;
	private static float				p01						= 0;
	private static float				p20						= 0;
	private static float				p11						= 0;
	private static float				p02						= 0;
	
	private static boolean			movementDataLoaded	= false;
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// -------------------------------------------------------------------------
	
	
	@SuppressWarnings("unused")
	private TrackedBall()
	{
		super(new BallID(), 0);
	}
	
	
	/**
	 * Providing a <strong>hard, deep</strong> copy of original
	 * 
	 * @param original
	 */
	public TrackedBall(final TrackedBall original)
	{
		super(original);
		onCam = original.onCam;
		pos = new Vector3(original.pos);
		vel = new Vector3(original.vel);
		acc = new Vector3(original.acc);
	}
	
	
	/**
	 * @param pos
	 * @param vel
	 * @param acc
	 * @param confidence
	 * @param onCam
	 */
	public TrackedBall(final IVector3 pos, final IVector3 vel, final IVector3 acc, final float confidence,
			final boolean onCam)
	{
		super(new BallID(), confidence);
		this.pos = new Vector3(pos);
		this.vel = new Vector3(vel);
		this.acc = new Vector3(acc);
		this.onCam = onCam;
	}
	
	
	/**
	 * @param motion
	 * @return
	 */
	public static TrackedBall motionToTrackedBall(final BallMotionResult motion)
	{
		final float xPos = (float) (motion.x / WPConfig.FILTER_CONVERT_MM_TO_INTERNAL_UNIT);
		final float xVel = (float) (motion.vx / WPConfig.FILTER_CONVERT_MperS_TO_INTERNAL_V);
		final float xAcc = (float) (motion.ax / WPConfig.FILTER_CONVERT_MperSS_TO_INTERNAL_A);
		final float yPos = (float) (motion.y / WPConfig.FILTER_CONVERT_MM_TO_INTERNAL_UNIT);
		final float yVel = (float) (motion.vy / WPConfig.FILTER_CONVERT_MperS_TO_INTERNAL_V);
		final float yAcc = (float) (motion.ay / WPConfig.FILTER_CONVERT_MperSS_TO_INTERNAL_A);
		final float zPos = (float) (motion.z / WPConfig.FILTER_CONVERT_MM_TO_INTERNAL_UNIT);
		final float zVel = (float) (motion.vz / WPConfig.FILTER_CONVERT_MperS_TO_INTERNAL_V);
		final float zAcc = (float) (motion.az / WPConfig.FILTER_CONVERT_MperSS_TO_INTERNAL_A);
		final float confidence = (float) motion.confidence;
		final boolean isOnCam = motion.onCam;
		
		IVector3 pos = new Vector3f(xPos, yPos, zPos);
		IVector3 vel = new Vector3f(xVel, yVel, zVel);
		IVector3 acc = new Vector3f(xAcc, yAcc, zAcc);
		return new TrackedBall(pos, vel, acc, confidence, isOnCam);
	}
	
	
	/**
	 * Mirror position, velocity and acceleration over x and y axis.
	 * Do NEVER call this in the AI!
	 */
	public void mirror()
	{
		pos.set(new Vector3(-pos.x(), -pos.y(), pos.z()));
		vel.set(new Vector3(-vel.x(), -vel.y(), vel.z()));
		acc.set(new Vector3(-acc.x(), -acc.y(), acc.z()));
	}
	
	
	/**
	 * gets the theoretical position of the ball after a given time
	 * 
	 * @param time [s]
	 * @return the position
	 */
	public IVector2 getPosByTime(final float time)
	{
		if (!movementDataLoaded)
		{
			log.warn("No ball movement data loaded!");
		}
		float x = getVel().getLength2() * 1000;
		float y = time * 1e3f;
		
		float result = p00 + (p10 * x) + (p01 * y) + (p20 * x * x) + (p11 * x * y) + (p02 * y * y);
		if (result < 0)
		{
			return getPos();
		}
		return getPos().addNew(getVel().scaleToNew(result));
	}
	
	
	/**
	 * @param time
	 * @param vel
	 * @return
	 */
	public IVector2 getPosByTime(final float time, final float vel)
	{
		if (!movementDataLoaded)
		{
			log.warn("No ball movement data loaded!");
		}
		float x = vel * 1000;
		float y = time * 1e3f;
		
		float result = p00 + (p10 * x) + (p01 * y) + (p20 * x * x) + (p11 * x * y) + (p02 * y * y);
		if (result < 0)
		{
			return getPos();
		}
		return getPos().addNew(getVel().scaleToNew(result));
	}
	
	
	/**
	 * gets the theoretical position of the ball when it reaches a given velocity
	 * 
	 * @param velocity [m/s]
	 * @return the position
	 */
	public IVector2 getPosByVel(final float velocity)
	{
		return getPosByTime(getTimeByVel(velocity));
	}
	
	
	/**
	 * gets the theoretical needed time for the ball to reach pos,
	 * pos should be on the balls movementPath.
	 * 
	 * @param pos
	 * @return the time [s]
	 */
	public float getTimeByPos(final IVector2 pos)
	{
		if (!movementDataLoaded)
		{
			log.warn("No ball movement data loaded!");
		}
		float dist = getPos().subtractNew(pos).getLength2();
		float x = getVel().getLength2() * 1000;
		float y = (p01 + (p11 * x) + SumatraMath.sqrt(SumatraMath.square(p01 + (p11 * x))
				- ((-4) * p02 * (dist - (p00 + (p10 * x) + (p20 * x * x))))))
				/ ((-2) * p02);
		y *= 1e-3f;
		return y;
	}
	
	
	/**
	 * gets the theoretical time where the ball reaches a given velocity
	 * 
	 * @param velocity [m/s]
	 * @return the time [s]
	 */
	public float getTimeByVel(final float velocity)
	{
		if (!movementDataLoaded)
		{
			log.warn("No ball movement data loaded!");
		}
		float x = getVel().getLength2() * 1000;
		float y = (((velocity - (2f * p20 * x) - p01) + (p11 * x)) - p10) / (p11 + (2f * p02));
		y = y * 1e-3f;
		return y;
	}
	
	
	/**
	 * gets the theoretical velocity of the ball at a given position
	 * 
	 * @param pos
	 * @return the velocity [m/s]
	 */
	public float getVelByPos(final IVector2 pos)
	{
		return getVelByTime(getTimeByPos(pos));
	}
	
	
	/**
	 * gets the theoretical velocity of the ball after a given time
	 * 
	 * @param time [s]
	 * @return the velocity [m/s]
	 */
	public float getVelByTime(final float time)
	{
		if (!movementDataLoaded)
		{
			log.warn("No ball movement data loaded!");
		}
		float x = getVel().getLength2() * 1000;
		float y = time * 1e3f;
		float velocity = ((((p11 * x) - ((2 * p20 * x) + p01) - p10) / (p11 + (2 * p02))) - y) * (p11 + (2 * p02));
		if (velocity < 0)
		{
			velocity = 0;
		}
		return velocity;
	}
	
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	
	@Override
	public BallID getId()
	{
		return (BallID) id;
	}
	
	
	/**
	 * @return
	 */
	public boolean isOnCam()
	{
		return onCam;
	}
	
	
	@Override
	public IVector2 getPos()
	{
		return new Vector2f(pos.x(), pos.y());
	}
	
	
	@Override
	public IVector2 getVel()
	{
		return new Vector2f(vel.x(), vel.y());
	}
	
	
	@Override
	public IVector2 getAcc()
	{
		return new Vector2f(acc.x(), acc.y());
	}
	
	
	/**
	 * @return
	 */
	public IVector3 getPos3()
	{
		return pos;
	}
	
	
	/**
	 * @return
	 */
	public IVector3 getVel3()
	{
		return vel;
	}
	
	
	/**
	 * @return
	 */
	public IVector3 getAcc3()
	{
		return acc;
	}
	
	
	/**
	 * @param pp00
	 * @param pp10
	 * @param pp01
	 * @param pp20
	 * @param pp11
	 * @param pp02
	 */
	public static void setBallMovementParameters(final float pp00, final float pp10, final float pp01, final float pp20,
			final float pp11, final float pp02)
	{
		p00 = pp00;
		p10 = pp10;
		p01 = pp01;
		p20 = pp20;
		p11 = pp11;
		p02 = pp02;
		log.debug("new ball parameters have been set: " + "p00: " + p00 + " p10: " + p10 +
				" p01: " + p01 + " p20: " + p20 + " p11: " + p11 + " p02: " + p02);
		
		String oldGrSimData = getBallMovementParameterLine("grSim");
		String oldRealData = getBallMovementParameterLine("real");
		
		String pathSep = System.getProperty("file.separator");
		String moduli = SumatraModel.getInstance().getCurrentModuliConfig().split("\\.")[0];
		String configPath = SumatraModel.getInstance().getUserPropertiesFile().getAbsolutePath().split("\\.")[0]
				+ "config" + pathSep + "generated" + pathSep + "ballMovement.config";
		
		Path path = FileSystems.getDefault().getPath(configPath);
		try (BufferedWriter bw = Files.newBufferedWriter(path, StandardCharsets.UTF_8))
		{
			String text = "";
			if (moduli.equals("moduli_sim"))
			{
				text = "grSim: " + p00 + "," + p10 + "," + p01
						+ "," + p20 + "," + p11 + "," + p02 + "\n" + oldRealData;
			} else
			{
				text = "real: " + p00 + "," + p10 + "," + p01
						+ "," + p20 + "," + p11 + "," + p02 + "\n" + oldGrSimData;
			}
			bw.write(text);
		} catch (IOException e)
		{
			log.error("Fatal error, couldn't write configFile: " + configPath);
		}
		movementDataLoaded = true;
	}
	
	
	/**
	 * loads ballMovement parameters generated by the
	 * BallMovementAnalyser
	 */
	public static void loadBallMovementParameters()
	{
		String moduli = SumatraModel.getInstance().getCurrentModuliConfig().split("\\.")[0];
		String parameters = "";
		if (moduli.equals("moduli_sim"))
		{
			parameters = getBallMovementParameterLine("grSim");
		} else
		{
			parameters = getBallMovementParameterLine("real");
		}
		parameters = parameters.split(":")[1];
		String[] parametersList = parameters.split(",");
		for (int i = 0; i < parametersList.length; i++)
		{
			if (i == 0)
			{
				p00 = Float.parseFloat(parametersList[i]);
			} else if (i == 1)
			{
				p10 = Float.parseFloat(parametersList[i]);
			} else if (i == 2)
			{
				p01 = Float.parseFloat(parametersList[i]);
			} else if (i == 3)
			{
				p20 = Float.parseFloat(parametersList[i]);
			} else if (i == 4)
			{
				p11 = Float.parseFloat(parametersList[i]);
			} else if (i == 5)
			{
				p02 = Float.parseFloat(parametersList[i]);
			}
		}
		
		movementDataLoaded = true;
	}
	
	
	private static String getBallMovementParameterLine(final String identifier)
	{
		String pathSep = System.getProperty("file.separator");
		String configPath = "config" + pathSep + "generated" + pathSep + "ballMovement.config";
		String grSimLine = "";
		String realLine = "";
		
		Path path = FileSystems.getDefault().getPath(configPath);
		try (BufferedReader in = Files.newBufferedReader(path, StandardCharsets.UTF_8))
		{
			String line = null;
			while ((line = in.readLine()) != null)
			{
				if (line.split(":")[0].equals("grSim"))
				{
					grSimLine = line;
				} else if (line.split(":")[0].equals("real"))
				{
					realLine = line;
				}
			}
		} catch (IOException e)
		{
			log.error("TrackedBall: fatal error while reading config File: " + configPath, e);
		}
		
		if (identifier.equals("grSim"))
		{
			return grSimLine;
		} else if (identifier.equals("real"))
		{
			return realLine;
		} else
		{
			log.error("TrackedBall: unknown identifier!");
			return null;
		}
	}
}
