/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Oct 7, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.general;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.BaseAiFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.EGameState;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.OffensiveStrategy;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.TacticalField;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.lachesis.RoleFinderInfo;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.ACalculator;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.plays.EPlay;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * Find number of roles and preferred bots for all roles
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class RoleFinderCalc extends ACalculator
{
	private static final Logger	log				= Logger.getLogger(RoleFinderCalc.class.getName());
	private boolean					noKeeperWarned	= true;
	
	
	@Configurable(comment = "numDesired defender bots")
	private static int				numDefenders	= 3;
	
	
	/**
	 * 
	 */
	public RoleFinderCalc()
	{
	}
	
	
	@Override
	public void doCalc(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		switch (newTacticalField.getGameState())
		{
			case PREPARE_PENALTY_THEY:
				keeper(newTacticalField, baseAiFrame);
				penaltyTheyMode(newTacticalField, baseAiFrame);
				break;
			case PREPARE_PENALTY_WE:
				keeper(newTacticalField, baseAiFrame);
				penaltyWeMode(newTacticalField, baseAiFrame);
				break;
			case BREAK:
			case TIMEOUT_THEY:
			case TIMEOUT_WE:
				breakGame(newTacticalField, baseAiFrame);
				break;
			case HALTED:
			case UNKNOWN:
				// no plays
				break;
			case POST_GAME:
				postGame(newTacticalField, baseAiFrame);
				break;
			// case STOPPED:
			// normalMode(newTacticalField, baseAiFrame);
			// for (RoleFinderInfo info : newTacticalField.getRoleFinderInfos().values())
			// {
			// info.getDesiredBots().clear();
			// }
			// keeper(newTacticalField, baseAiFrame);
			// break;
			default:
				keeper(newTacticalField, baseAiFrame);
				normalMode(newTacticalField, baseAiFrame);
				break;
		}
	}
	
	
	private void breakGame(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		int numRoles = baseAiFrame.getWorldFrame().getTigerBotsAvailable().size();
		newTacticalField.getRoleFinderInfos().put(EPlay.MAINTENANCE, new RoleFinderInfo(0, numRoles, numRoles));
	}
	
	
	private void postGame(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		int numRoles = baseAiFrame.getWorldFrame().getTigerBotsAvailable().size();
		newTacticalField.getRoleFinderInfos().put(EPlay.CHEERING, new RoleFinderInfo(0, numRoles, numRoles));
	}
	
	
	/**
	 * @param newTacticalField
	 * @param baseAiFrame
	 */
	private void penaltyWeMode(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		int numRoles = Math.max(1, baseAiFrame.getWorldFrame().getTigerBotsAvailable().size() - 3);
		normalDefense(newTacticalField, baseAiFrame);
		newTacticalField.getRoleFinderInfos().put(EPlay.PENALTY_WE, new RoleFinderInfo(0, numRoles, numRoles));
	}
	
	
	/**
	 * @param newTacticalField
	 * @param baseAiFrame
	 */
	private void penaltyTheyMode(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		newTacticalField.getRoleFinderInfos().put(EPlay.PENALTY_THEM, new RoleFinderInfo(0, 10, 5));
	}
	
	
	private void keeper(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		final RoleFinderInfo keeperInfo;
		// if keeper is present request it, else request no, because it is against the law to enter penArea with any other
		// id
		if (baseAiFrame.getWorldFrame().getTigerBotsAvailable().containsKey(baseAiFrame.getKeeperId()))
		{
			keeperInfo = new RoleFinderInfo(1, 1, 1);
			keeperInfo.getDesiredBots().add(baseAiFrame.getKeeperId());
			keeperInfo.setForceNumDesiredBots(1);
			noKeeperWarned = false;
		} else
		{
			if (!noKeeperWarned)
			{
				log.warn("Our keeper id (" + baseAiFrame.getKeeperId() + ") is not present!");
			}
			noKeeperWarned = true;
			keeperInfo = new RoleFinderInfo(0, 0, 0);
		}
		newTacticalField.getRoleFinderInfos().put(EPlay.KEEPER, keeperInfo);
	}
	
	
	private void normalDefense(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		int maxDefs = newTacticalField.getDangerousFoeBots().size() + 2;
		int maxRoles = Math.min(maxDefs, numDefenders);
		int desRoles = Math.min(maxDefs, getNumDefenders(newTacticalField, baseAiFrame));
		
		RoleFinderInfo defenderInfo = new RoleFinderInfo(0, maxRoles, desRoles);
		List<TrackedTigerBot> desiredBots = new ArrayList<TrackedTigerBot>(baseAiFrame.getWorldFrame()
				.getTigerBotsAvailable().values());
		Collections.sort(desiredBots, TrackedTigerBot.DISTANCE_TO_GOAL_COMPARATOR);
		for (TrackedTigerBot tiger : desiredBots)
		{
			defenderInfo.getDesiredBots().add(tiger.getId());
		}
		
		newTacticalField.getRoleFinderInfos().put(EPlay.DEFENSIVE, defenderInfo);
	}
	
	
	private void normalMode(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		normalDefense(newTacticalField, baseAiFrame);
		
		OffensiveStrategy offensiveStrategy = newTacticalField.getOffensiveStrategy();
		int desiredBots;
		if (offensiveStrategy.getDesiredBots() == null)
		{
			desiredBots = 0;
		} else
		{
			desiredBots = offensiveStrategy.getDesiredBots().size();
		}
		// desiredBots = desiredBots + offensiveStrategy.getHelperDestinations().size();
		RoleFinderInfo offenceInfo = new RoleFinderInfo(
				offensiveStrategy.getMinNumberOfBots(),
				offensiveStrategy.getMaxNumberOfBots(),
				desiredBots);
		if (offensiveStrategy.getDesiredBots() != null)
		{
			for (BotID key : offensiveStrategy.getDesiredBots())
			{
				if (key != null)
				{
					offenceInfo.getDesiredBots().add(key);
				}
			}
		}
		newTacticalField.getRoleFinderInfos().put(EPlay.OFFENSIVE, offenceInfo);
		
		RoleFinderInfo supporterInfo = new RoleFinderInfo(0, 6, 2);
		newTacticalField.getRoleFinderInfos().put(EPlay.SUPPORT, supporterInfo);
	}
	
	
	private int getNumDefenders(final TacticalField tacticalField, final BaseAiFrame aiFrame)
	{
		int numDefenders = RoleFinderCalc.numDefenders;
		int goalsThem = aiFrame.getLatestRefereeMsg().getTeamInfoThem().getScore();
		int goalsUs = aiFrame.getLatestRefereeMsg().getTeamInfoTigers().getScore();
		
		if (goalsThem < goalsUs)
		{
			numDefenders++;
		} else if (goalsThem > goalsUs)
		{
			numDefenders--;
		}
		// in stop predict next GameState
		EGameState state = tacticalField.getGameState();
		if (state == EGameState.STOPPED)
		{
			// we touched the ball last
			if (aiFrame.getWorldFrame().getTigerBotsAvailable().containsKey(tacticalField.getBotLastTouchedBall()))
			{
				numDefenders++;
			} else
			{
				numDefenders--;
			}
		} else
		{
			numDefenders += isDangerousSituation(state, aiFrame) ? 1 : 0;
			numDefenders -= isOffensiveSituation(state, aiFrame) ? 1 : 0;
		}
		
		return Math.max(numDefenders, 0);
	}
	
	
	private boolean isDangerousSituation(final EGameState gameState, final BaseAiFrame aiFrame)
	{
		if ((gameState == EGameState.GOAL_KICK_THEY)
				|| (gameState == EGameState.CORNER_KICK_THEY)
				|| (gameState == EGameState.DIRECT_KICK_THEY)
				|| (gameState == EGameState.THROW_IN_THEY))
		{
			return true;
		}
		// in stop predict the next GameState and set
		return false;
	}
	
	
	private boolean isOffensiveSituation(final EGameState gameState, final BaseAiFrame aiFrame)
	{
		if ((gameState == EGameState.CORNER_KICK_WE)
				|| (gameState == EGameState.PREPARE_KICKOFF_WE)
				|| (gameState == EGameState.DIRECT_KICK_WE)
				|| (gameState == EGameState.THROW_IN_WE))
		{
			return true;
		}
		return false;
	}
}
