/*
 * *********************************************************
 * Copyright (c) 2009 - 2013, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: 04.04.2015
 * Author(s): MarkG
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.offense;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import edu.dhbw.mannheim.tigers.sumatra.model.data.DynamicPosition;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.BaseAiFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AiMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.EGameState;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.TacticalField;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.AIConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.ACalculator;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.offense.data.OffensiveAction;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.offense.data.OffensiveAction.EOffensiveAction;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.support.data.AdvancedPassTarget;


/**
 * Calculates offensive Actions for the OffenseRole.
 * 
 * @author Mark Geiger <Mark.Geiger@dlr.de>
 */
public class OffensiveActionsCalc extends ACalculator
{
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// -------------------------------------- ------------------------------------
	
	
	// --------------------------------------------------------------------------
	// --- methods --------------------------------------------------------------
	// --------------------------------------------------------------------------
	
	@Override
	public void doCalc(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		Map<BotID, OffensiveAction> offensiveActions = new HashMap<BotID, OffensiveAction>();
		for (BotID key : baseAiFrame.getWorldFrame().getTigerBotsAvailable().keySet())
		{
			offensiveActions.put(key, calcOffensiveAction(key, newTacticalField, baseAiFrame));
		}
		newTacticalField.setOffensiveActions(offensiveActions);
	}
	
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	
	private OffensiveAction calcOffensiveAction(final BotID botID, final TacticalField newTacticalField,
			final BaseAiFrame baseAiFrame)
	{
		
		OffensiveAction action = new OffensiveAction();
		
		if (newTacticalField.getGameState() == EGameState.PREPARE_KICKOFF_WE) // Kickoff
		{
			action.setType(EOffensiveAction.GOAL_SHOT);
			action.setDirectShotAndClearingTarget(getBestShootTarget(newTacticalField));
		} else if (isDirectGoalShootPossible(newTacticalField,
				baseAiFrame) /* && !(/* forcePass && OffensiveConstants.forcePassWhenIndirectIsCalled */)
		{
			action.setType(EOffensiveAction.GOAL_SHOT);
			action.setDirectShotAndClearingTarget(getBestShootTarget(newTacticalField));
		} else if (setPassParamsIfPassIsPossible(botID, newTacticalField,
				baseAiFrame) /* forcePass || *//* OffensiveConstants.forcePassWhenIndirectIsCalled */)
		{
			action.setType(EOffensiveAction.PASS);
			action.setPassTarget(selectPassTarget(botID, newTacticalField, baseAiFrame));
		} else if (isLowScoringChanceDirectGoalShootPossible(newTacticalField,
				baseAiFrame))
		{
			if (isBotInGoodBlockingPosition(botID, newTacticalField, baseAiFrame))
			{
				action.setType(EOffensiveAction.PASS);
				action.setPassTarget(selectPassTarget(botID, newTacticalField, baseAiFrame));
			}
			else
			{
				action.setType(EOffensiveAction.GOAL_SHOT);
				action.setDirectShotAndClearingTarget(getBestShootTarget(newTacticalField));
			}
		}
		return action;
	}
	
	
	/**
	 * selectPassTarget:
	 * determines which target to pass to. Will throw Excepion if no pass to any target is possible.
	 * 
	 * @param botID
	 * @param newTacticalField
	 * @param baseAiFrame
	 * @return
	 */
	
	private AdvancedPassTarget selectPassTarget(final BotID botID, final TacticalField newTacticalField,
			final BaseAiFrame baseAiFrame)
	{
		List<AdvancedPassTarget> advancedPassTargets = newTacticalField.getAdvancedPassTargetsRanked();
		for (AdvancedPassTarget key : advancedPassTargets)
		{
			BotID passTarget = key.getBotId();
			float distance = GeoMath.distancePP(baseAiFrame.getWorldFrame().getBot(botID).getPos(), key);
			if ((passTarget != botID) && (distance > 1500f) && isPassTargetReachable(baseAiFrame, botID, key, passTarget))
			{
				
				return new AdvancedPassTarget(key, key.value, false, key.getBotId());
			}
		}
		
		for (AdvancedPassTarget key : advancedPassTargets)
		{
			BotID passTarget = key.getBotId();
			float distance = GeoMath.distancePP(baseAiFrame.getWorldFrame().getBot(botID).getPos(), key);
			if ((passTarget != botID) && (distance > 2000f))
			{
				return new AdvancedPassTarget(key, key.value, true, key.getBotId());
			}
		}
		return advancedPassTargets.get(0); // This line won't be reached.
	}
	
	
	/*
	 * isBotInGoodBlockingPosition:
	 * this function checks whether current bot is in good position to prevent foe from goal_kick
	 * (maybe then current tiger should not move away for goal_kick if chance is rather low)
	 */
	private boolean isBotInGoodBlockingPosition(final BotID botID, final TacticalField currentTacticalField,
			final BaseAiFrame baseAiFrame)
	{
		BotID ballPossessorsID = currentTacticalField.getBallPossession().getOpponentsId();
		
		// Create List of all Tigers, except botID
		List<BotID> botstoignore = new ArrayList<BotID>();
		for (Entry<BotID, TrackedTigerBot> tiger : baseAiFrame.getWorldFrame().tigerBotsVisible)
		{
			if (tiger.getKey() == botID)
			{
				continue;
			}
			botstoignore.add(tiger.getKey());
		}
		// //
		float raysize = AIConfig.getGeometry().getBotRadius() * 2f;
		try
		{
			IVector2 ourGoalPosition = AIConfig.getGeometry().getGoalTheir().getGoalCenter();
			
			IVector2 ballPossessorsPos = baseAiFrame.getWorldFrame().getBot(ballPossessorsID).getPos();
			return !GeoMath.p2pVisibility(baseAiFrame.getWorldFrame(), ballPossessorsPos, ourGoalPosition, raysize,
					botstoignore);
		} catch (Exception e)
		{
			return false;
		}
		
	}
	
	
	private boolean isPassTargetReachable(final BaseAiFrame baseAiFrame, final BotID thisbot,
			final AdvancedPassTarget target, final BotID targetbot)
	{
		IVector2 startPos = baseAiFrame.getWorldFrame().getBot(thisbot).getPos();
		List<BotID> ignoredBots = new ArrayList<BotID>(2);
		ignoredBots.add(thisbot);
		ignoredBots.add(targetbot);
		return GeoMath.p2pVisibility(baseAiFrame.getWorldFrame(), startPos, target, ignoredBots);
	}
	
	
	private boolean isDirectGoalShootPossible(final TacticalField newTacticalField, final BaseAiFrame baseAiFrame)
	{
		if (AiMath.willBotShoot(baseAiFrame.getWorldFrame(), false)
				&& (newTacticalField.getBestDirectShootTarget() != null))
		{
			return true;
		}
		return false;
	}
	
	
	private boolean setPassParamsIfPassIsPossible(final BotID id, final TacticalField newTacticalField,
			final BaseAiFrame baseAiFrame)
	{
		IVector2 helperPos = null;
		BotID passTarget = null;
		
		List<AdvancedPassTarget> advancedPassTargets = newTacticalField.getAdvancedPassTargetsRanked();
		for (AdvancedPassTarget key : advancedPassTargets)
		{
			helperPos = key;
			passTarget = key.getBotId();
			float distance = GeoMath.distancePP(baseAiFrame.getWorldFrame().getBot(id).getPos(), key);
			if ((passTarget != id) && (distance > 1500f) && (isPassTargetReachable(baseAiFrame, id, key, passTarget)))
			{
				break;
			}
		}
		
		if ((helperPos != null) && (passTarget != null) && (passTarget != id))
		{
			return true;
		}
		
		for (AdvancedPassTarget key : advancedPassTargets)
		{
			helperPos = key;
			passTarget = key.getBotId();
			float distance = GeoMath.distancePP(baseAiFrame.getWorldFrame().getBot(id).getPos(), key);
			if ((passTarget != id) && (distance > 2000f))
			{
				return true;
			}
		}
		
		
		return false;
	}
	
	
	private boolean isLowScoringChanceDirectGoalShootPossible(final TacticalField newTacticalField,
			final BaseAiFrame baseAiFrame)
	{
		if (newTacticalField.getBestDirectShootTarget() != null)
		{
			return true;
		}
		return false;
	}
	
	
	private DynamicPosition getBestShootTarget(final TacticalField newTacticalField)
	{
		IVector2 target = newTacticalField.getBestDirectShootTarget();
		if (target == null)
		{
			target = AIConfig.getGeometry().getGoalTheir().getGoalCenter();
		}
		return new DynamicPosition(target);
	}
}
