/*
 * *********************************************************
 * Copyright (c) 2009 - 2015, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Apr 16, 2015
 * Author(s): FelixB <bayer.fel@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.plays.defense.algorithms;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.MetisAiFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.exceptions.MathException;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.EGameState;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.circle.Circle;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.circle.DrawableCircle;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.AVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.line.DrawableLine;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.line.Line;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.AIConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.Geometry;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.defense.data.DefensePoint;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.metis.calculators.defense.data.FoeBotData;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.plays.defense.algorithms.interfaces.IDefensePointCalc;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.defense.DefenderRole;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;
import edu.dhbw.mannheim.tigers.sumatra.view.visualizer.internals.field.EDrawableShapesLayer;


/**
 * This calculator aims to create a defensive pattern which uses two bots to defense versus the ball and a single
 * bot for every other enemy offensive bot.
 * 
 * @author FelixB <bayer.fel@gmail.com>
 */
public class DoubleDefPointCalcV2 implements IDefensePointCalc
{
	
	
	private static final Logger		log								= Logger.getLogger(DoubleDefPointCalcV2.class.getName());
	
	@Configurable(comment = "Time [s] - the defenders who cover the ball will try to cover the future ball")
	private static float					ballLookaheadOfDefenders	= 1f;
	
	private static float					botRadius						= AIConfig.getGeometry().getBotRadius();
	
	private static float					penAreaMargin					= Geometry.getPenaltyAreaMargin();
	
	@Configurable(comment = "Defenders will not drive behind this x value")
	private static float					maxXDefender					= -1500;
	
	@Configurable(comment = "Minimum distance of the double block to the ball")
	private static float					minDistDefenderBall			= 1.5f * botRadius;
	
	
	private final IDefensePointCalc	parentCalc;
	
	
	/**
	 * @param parentCalc
	 */
	public DoubleDefPointCalcV2(final IDefensePointCalc parentCalc)
	{
		this.parentCalc = parentCalc;
	}
	
	
	@Override
	public Map<DefenderRole, DefensePoint> getDefenderDistribution(final MetisAiFrame frame,
			final List<DefenderRole> defenders,
			final List<FoeBotData> foeBotDataList)
	{
		
		Map<DefenderRole, DefensePoint> optDistr = new HashMap<DefenderRole, DefensePoint>();
		if (defenders.isEmpty())
		{
			return optDistr;
		}
		
		getDirectShotDistr(frame, defenders, optDistr);
		for (FoeBotData curFoeBot : new ArrayList<FoeBotData>(foeBotDataList))
		{
			if (curFoeBot.posessesBall())
			{
				foeBotDataList.remove(curFoeBot);
			}
		}
		
		List<DefenderRole> remainingDefenders = new ArrayList<DefenderRole>();
		
		remainingDefenders.addAll(defenders.stream().filter(el -> !optDistr.keySet().contains(el))
				.collect(Collectors.toList()));
		
		optDistr.putAll(parentCalc.getDefenderDistribution(frame, remainingDefenders, foeBotDataList));
		
		return optDistr;
	}
	
	
	private void getDirectShotDistr(final MetisAiFrame frame, final List<DefenderRole> defenders,
			final Map<DefenderRole, DefensePoint> optDistr)
	{
		
		IVector2 ballPos = frame.getWorldFrame().getBall().getPosByTime(ballLookaheadOfDefenders);
		ballPos = AIConfig.getGeometry().getField().nearestPointInside(ballPos, AIConfig.getGeometry().getBallRadius());
		float width;
		
		IVector2 intersectionBisectorGoal = GeoMath.calculateBisector(ballPos, AIConfig.getGeometry().getGoalOur()
				.getGoalPostLeft(), AIConfig.getGeometry().getGoalOur().getGoalPostRight());
		
		float angleBallLeftGoal = GeoMath.angleBetweenVectorAndVector(AIConfig.getGeometry().getGoalOur()
				.getGoalPostLeft().subtractNew(ballPos), intersectionBisectorGoal.subtractNew(ballPos));
		
		
		if ((1 == defenders.size()) || !frame.getTacticalField().needTwoForBallBlock())
		{
			width = botRadius;
		} else
		{
			width = 2f * botRadius;
		}
		
		float distBall2DefPoint = width / AngleMath.tan(angleBallLeftGoal);
		distBall2DefPoint = Math.min(distBall2DefPoint, GeoMath.distancePP(ballPos, intersectionBisectorGoal));
		
		// calculate the valid range on the intersection
		IVector2 minXOnIntersection = AIConfig.getGeometry().getPenaltyAreaOur()
				.nearestPointOutside(intersectionBisectorGoal, ballPos, penAreaMargin);
		IVector2 maxXOnIntersection = AIConfig.getGeometry().getPenaltyAreaOur()
				.nearestPointOutside(GeoMath.stepAlongLine(ballPos, intersectionBisectorGoal, minDistDefenderBall));
		maxXOnIntersection = limitX(maxXOnIntersection, intersectionBisectorGoal);
		if (EGameState.STOPPED == frame.getTacticalField().getGameState())
		{
			maxXOnIntersection = modifyForStopSituation(maxXOnIntersection, ballPos, intersectionBisectorGoal);
		}
		maxXOnIntersection = AIConfig.getGeometry().getPenaltyAreaOur()
				.nearestPointOutside(maxXOnIntersection, ballPos, penAreaMargin);
		
		// Create the defense point on the bisector and assure it lies in a valid range
		IVector2 defPoint = GeoMath.stepAlongLine(ballPos, intersectionBisectorGoal, distBall2DefPoint);
		
		if (!GeoMath.isVectorBetween(defPoint, minXOnIntersection, maxXOnIntersection))
		{
			if (GeoMath.distancePP(defPoint, minXOnIntersection) < GeoMath.distancePP(defPoint, maxXOnIntersection))
			{
				defPoint = minXOnIntersection;
			} else
			{
				defPoint = maxXOnIntersection;
			}
		}
		
		float triangleLength = GeoMath.triangleDistance(frame.getWorldFrame().getBall().getPos(), AIConfig.getGeometry()
				.getGoalOur()
				.getGoalPostLeft(), AIConfig.getGeometry().getGoalOur()
				.getGoalPostRight(), defPoint);
		
		if (1 == defenders.size())
		{
			optDistr.put(defenders.get(0), new DefensePoint(defPoint));
		} else if (triangleLength < (AIConfig.getGeometry().getBotRadius() * 2))// !frame.getTacticalField().needTwoForBallBlock())
		{
			optDistr.put(getOptimalSingleDefender(defPoint, defenders), new DefensePoint(defPoint));
		} else
		{
			optDistr.putAll(getOptimalDoubleDefenders(defPoint, intersectionBisectorGoal, defenders,
					frame));
		}
	}
	
	
	private DefenderRole getOptimalSingleDefender(final IVector2 defPoint, final List<DefenderRole> defenders)
	{
		DefenderRole optDefender = null;
		float optDefenderDist = Float.MAX_VALUE;
		float curDistance;
		
		for (DefenderRole curDefender : defenders)
		{
			curDistance = GeoMath.distancePP(curDefender.getPos(), defPoint);
			if (curDistance < optDefenderDist)
			{
				optDefender = curDefender;
				optDefenderDist = curDistance;
			}
		}
		
		return optDefender;
	}
	
	
	private Map<DefenderRole, DefensePoint> getOptimalDoubleDefenders(final IVector2 defPoint,
			final IVector2 bisectorBall2Posts, final List<DefenderRole> defenders, final MetisAiFrame frame)
	{
		// code by Dirk: bots should be rather covering the goal posts than the goal center if they cannot cover the whole
		// goal
		IVector2 ballPos = frame.getWorldFrame().getBall().getPos();
		frame.getTacticalField()
				.getDrawableShapes()
				.get(EDrawableShapesLayer.DEFENSE)
				.add(new DrawableLine(Line.newLine(ballPos,
						bisectorBall2Posts),
						Color.PINK));
		
		
		float defenseLineLength = GeoMath.triangleDistance(ballPos, AIConfig.getGeometry().getGoalOur()
				.getGoalPostLeft(), AIConfig.getGeometry().getGoalOur()
				.getGoalPostRight(), defPoint);
		
		float distanceToDefPoint = Math.max((defenseLineLength / 2) - AIConfig.getGeometry().getBotRadius(), AIConfig
				.getGeometry()
				.getBotRadius());
		
		List<IVector2> defensePoints = new ArrayList<IVector2>();
		
		frame.getTacticalField()
				.getDrawableShapes()
				.get(EDrawableShapesLayer.DEFENSE)
				.add(new DrawableLine(Line.newLine(frame.getWorldFrame().getBall().getPos(),
						bisectorBall2Posts),
						Color.pink));
		
		defensePoints.add(defPoint.addNew(ballPos.subtractNew(bisectorBall2Posts).scaleToNew(distanceToDefPoint)
				.turnNew(-AngleMath.PI_HALF)));
		defensePoints.add(defPoint.addNew(ballPos.subtractNew(bisectorBall2Posts).scaleToNew(distanceToDefPoint).turnNew(
				AngleMath.PI_HALF)));
		frame.getTacticalField()
				.getDrawableShapes()
				.get(EDrawableShapesLayer.DEFENSE)
				.add(new DrawableLine(Line.newLine(defPoint,
						defensePoints.get(0)),
						Color.PINK));
		frame.getTacticalField()
				.getDrawableShapes()
				.get(EDrawableShapesLayer.DEFENSE)
				.add(new DrawableLine(Line.newLine(defPoint,
						defensePoints.get(1)),
						Color.PINK));
		// end: code by dirk
		
		// get the defenders driving the shortest accumulated distance
		// dirk: this does not seem to work, the defenders cross their pathes when they drive through y=0
		float curOptDist = Float.MAX_VALUE;
		float curDist = Float.MAX_VALUE;
		// Map<DefenderRole, DefensePoint> curOptDistr = new HashMap<DefenderRole, DefensePoint>();
		List<DefenderRole> closestDefender = new ArrayList<DefenderRole>();
		for (DefenderRole defender1 : defenders.subList(0, defenders.size() - 1))
		{
			for (DefenderRole defender2 : defenders.subList(defenders.indexOf(defender1) + 1, defenders.size()))
			{
				curDist = GeoMath.distancePP(defender1.getPos(), defensePoints.get(0))
						+ GeoMath.distancePP(defender2.getPos(), defensePoints.get(1));
				if (curDist < curOptDist)
				{
					closestDefender.clear();
					closestDefender.add(defender1);
					closestDefender.add(defender2);
					curOptDist = curDist;
				}
				
				curDist = GeoMath.distancePP(defender2.getPos(), defensePoints.get(0))
						+ GeoMath.distancePP(defender1.getPos(), defensePoints.get(1));
				if (curDist < curOptDist)
				{
					closestDefender.clear();
					closestDefender.add(defender2);
					closestDefender.add(defender1);
					curOptDist = curDist;
				}
			}
		}
		
		// dirk: approach: sort both closest defenders by Y value:
		Map<DefenderRole, DefensePoint> curOptDistr = new HashMap<DefenderRole, DefensePoint>();
		
		closestDefender.sort(DefenderRole.Y_COMPARATOR);
		defensePoints.sort(Vector2.Y_COMPARATOR);
		
		// special case, one bot is too far away from the defense points, e.g. one of the defenders just got an offensive,
		// now a new bot from the other side of the field needs to drive to the defense. This takes some time. Meanwhile
		// the remaining bot NEEDS (it happens really often and is dangerous) to cover the more dangerous goal post!
		if ((GeoMath.distancePP(closestDefender.get(0).getPos(), defensePoints.get(0)) > (AIConfig.getGeometry()
				.getBotRadius() * 4))
				^
				(GeoMath.distancePP(closestDefender.get(1).getPos(), defensePoints.get(1)) > (AIConfig.getGeometry()
						.getBotRadius() * 4)))
		{
			DefenderRole closeByDefender = closestDefender.get(0);
			DefenderRole farAwayDefender = closestDefender.get(1);
			if (GeoMath.distancePP(closestDefender.get(1).getPos(), defensePoints.get(1)) < GeoMath.distancePP(
					closestDefender.get(0).getPos(), defensePoints.get(0)))
			{
				closeByDefender = closestDefender.get(1);
				farAwayDefender = closestDefender.get(0);
			}
			List<BotID> botsToIgnore = new ArrayList<BotID>();
			botsToIgnore.add(closeByDefender.getBotID());
			botsToIgnore.add(farAwayDefender.getBotID());
			float dangerLeft = GeoMath.p2pVisibilityVoting(frame.getWorldFrame(), ballPos, AIConfig.getGeometry()
					.getGoalOur().getGoalPostLeft(), AIConfig.getGeometry().getBotRadius(), botsToIgnore);
			float dangerRight = GeoMath.p2pVisibilityVoting(frame.getWorldFrame(), ballPos, AIConfig.getGeometry()
					.getGoalOur().getGoalPostRight(), AIConfig.getGeometry().getBotRadius(), botsToIgnore);
			closestDefender.clear();
			if (dangerLeft > dangerRight)
			{
				// the left post has always the higher y value
				closestDefender.add(closeByDefender);
				closestDefender.add(farAwayDefender);
			} else
			{
				closestDefender.add(farAwayDefender);
				closestDefender.add(closeByDefender);
			}
		}
		
		for (int i = 0; (i < defenders.size()) && (i < 2); i++)
		{
			// TODO: Felix, Handle maxX < minX
			IVector2 validDefPoint = AIConfig.getGeometry().getPenaltyAreaOur()
					.nearestPointOutside(defensePoints.get(i), ballPos, penAreaMargin);
			
			validDefPoint = AIConfig.getGeometry().getField().nearestPointInside(validDefPoint);
			curOptDistr.put(defenders.get(i), new DefensePoint(validDefPoint, frame.getWorldFrame().getBall()));
			frame.getTacticalField().getDrawableShapes().get(EDrawableShapesLayer.DEFENSE)
					.add(new DrawableCircle(defensePoints.get(i), 50, Color.cyan));
		}
		return curOptDistr;
	}
	
	
	/**
	 * TODO: implement a universal version for a point on a vector
	 * 
	 * @param point
	 * @return
	 */
	private IVector2 limitX(IVector2 point, final IVector2 intersectionBisectorGoal)
	{
		if (point.x() > maxXDefender)
		{
			try
			{
				IVector2 intersect = GeoMath.intersectionPoint(intersectionBisectorGoal,
						intersectionBisectorGoal.subtractNew(point), new Vector2(
								maxXDefender, 0), new Vector2(0, AVector2.Y_AXIS.y()));
				point = (
						(GeoMath.stepAlongLine(intersect, intersectionBisectorGoal,
								-AIConfig.getGeometry().getBotRadius())));
			} catch (MathException err)
			{
				// unlikely :P
				log.error("Unlikely error happended in DoubleDefPointCalcV2");
			}
		}
		return point;
	}
	
	
	private IVector2 modifyForStopSituation(final IVector2 dest, final IVector2 ballPos,
			final IVector2 intersectionBisectorGoal)
	{
		float bot2ballStopDist = AIConfig.getGeometry().getBotToBallDistanceStop();
		float botRadius = AIConfig.getGeometry().getBotRadius();
		
		Circle stopCircle = new Circle(ballPos, bot2ballStopDist);
		if (stopCircle.isPointInShape(dest, botRadius))
		{
			return GeoMath.stepAlongLine(ballPos, intersectionBisectorGoal, bot2ballStopDist + botRadius);
		}
		return dest;
	}
}
