/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Dec 13, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.test;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.ARole;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.ERole;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.EMovingSpeed;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.EFeature;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.AMoveSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.AMoveSkill.ECommandType;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.IMoveToSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.ISkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.PositionSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.statemachine.IRoleState;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;
import edu.dhbw.mannheim.tigers.sumatra.util.learning.GaussianPolicy;
import edu.dhbw.mannheim.tigers.sumatra.util.learning.PolicyParameters;


/**
 * Evaluate a learned policy
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class PolicyEvaluationRole extends ARole
{
	private static final Logger		log			= Logger.getLogger(PolicyEvaluationRole.class.getName());
	private final List<Path>			trialDirs;
	private int								trialNo		= 0;
	private final IVector3				dest;
	@Configurable(comment = "Precision [m] for reaching position")
	private static float					precision	= 0.001f;
	
	@Configurable(comment = "Time [s] after evaluation will be canceled")
	private static float					timeout		= 10f;
	
	@Configurable(comment = "Initial position before each evaluation")
	private static IVector2				initPos		= new Vector2(0, 0);
	
	private final List<EvalResult>	results		= new ArrayList<EvalResult>();
	
	private static class EvalResult
	{
		private int			evalId;
		private int			trialId;
		private float		time;
		private IVector3	offset;
		private float		offsetLen;
		private float		precision;
		private float		integral;
	}
	
	
	/**
	 * @param basePath
	 * @param dest
	 */
	public PolicyEvaluationRole(final String basePath, final IVector3 dest)
	{
		super(ERole.POLICY_EVALUATOR);
		this.dest = dest;
		
		Path path = Paths.get(basePath);
		trialDirs = listFiles(path).filter(p1 -> p1.getFileName().toString().startsWith("eval"))
				.flatMap(p ->
						listFiles(p).filter(p2 -> p2.getFileName().toString().startsWith("trial")))
				.sorted()
				.collect(Collectors.toList());
		
		IRoleState evalState = new EvaluationState();
		IRoleState initState = new InitState();
		setInitialState(initState);
		addTransition(EStateId.INIT, EEvent.INIT_DONE, evalState);
		addTransition(EStateId.EVAL, EEvent.EVAL_DONE, initState);
	}
	
	
	private static Stream<Path> listFiles(final Path p)
	{
		try
		{
			return Files.list(p);
		} catch (IOException err)
		{
			log.error("Could not list files.", err);
		}
		return Stream.of();
	}
	
	private enum EStateId
	{
		EVAL,
		INIT
	}
	
	private enum EEvent
	{
		INIT_DONE,
		EVAL_DONE
	}
	
	
	@Override
	public void fillNeededFeatures(final List<EFeature> features)
	{
	}
	
	private class EvaluationState implements IRoleState
	{
		private long					starttime;
		private long					lastTime;
		private EvalResult			result;
		private static final float	DT	= 0.1f;
		
		
		@Override
		public void doEntryActions()
		{
			Path lastFile = null;
			while ((trialNo) < trialDirs.size())
			{
				Path nextTrialPath = trialDirs.get(trialNo++);
				List<Path> controllerFiles = listFiles(nextTrialPath).filter(
						p -> p.getFileName().toString().startsWith("controller"))
						.sorted()
						.collect(Collectors.toList());
				if (controllerFiles.isEmpty())
				{
					log.warn("No controller file for " + nextTrialPath.toString());
				} else
				{
					lastFile = controllerFiles.get(controllerFiles.size() - 1);
					break;
				}
			}
			if (lastFile == null)
			{
				StringBuilder sb = new StringBuilder();
				for (EvalResult res : results)
				{
					sb.append(String.format("%n%d\t%d\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f\t%.5f", res.evalId, res.trialId,
							res.time,
							res.offsetLen, res.precision,
							res.offset.x(), res.offset.y(), res.offset.z()));
				}
				log.info(sb.toString());
				setCompleted();
				return;
			}
			PolicyParameters params;
			try
			{
				params = PolicyParameters.fromFile(lastFile.toString());
				GaussianPolicy gp = GaussianPolicy.fromParameters(params, "", false);
				PositionSkill skill = new PositionSkill(dest.getXYVector(), dest.z());
				skill.setGpController(gp);
				skill.setCommandType(ECommandType.POS_LEARNED);
				skill.getMoveCon().setSpeed(EMovingSpeed.UNLIMITED, 3);
				// skill.setPathDriver(new PositionDriver());
				setNewSkill(skill);
				starttime = SumatraClock.nanoTime();
				lastTime = 0;
				result = new EvalResult();
				Pattern p = Pattern.compile(".*/eval([0-9]*)/trial([0-9]*).*");
				Matcher m = p.matcher(lastFile.toString());
				if (m.matches())
				{
					String strEval = m.group(1);
					String strTrial = m.group(2);
					result.evalId = Integer.valueOf(strEval);
					result.trialId = Integer.valueOf(strTrial);
				}
				log.info("Start evaluating " + lastFile);
			} catch (IOException err)
			{
				log.error("Could not read policy parameters: " + lastFile.toString(), err);
				nextState(EEvent.EVAL_DONE);
			}
		}
		
		
		@Override
		public void doUpdate()
		{
			if ((SumatraClock.nanoTime() - lastTime) > (DT * 1e9))
			{
				lastTime = SumatraClock.nanoTime();
				result.integral += getOffset().getLength3();
			}
			
			float dt = (SumatraClock.nanoTime() - starttime) * 1e-9f;
			if (dt > timeout)
			{
				done("timeout");
			}
			
			if (getOffset().getLength3() < precision)
			{
				done("dest reached");
			}
		}
		
		
		@Override
		public void doExitActions()
		{
		}
		
		
		@Override
		public void onSkillStarted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public void onSkillCompleted(final ISkill skill, final BotID botID)
		{
		}
		
		
		/**
		 * @return offset in [m,m,rad]
		 */
		private IVector3 getOffset()
		{
			IVector3 pos = new Vector3(getPos(), getBot().getAngle());
			IVector3 offset = dest.subtractNew(pos);
			IVector3 offsetScaled = new Vector3(offset.getXYVector().multiply(1e-3f), AngleMath.normalizeAngle(offset.z()))
					.abs();
			return offsetScaled;
		}
		
		
		private void done(final String reason)
		{
			result.time = (SumatraClock.nanoTime() - starttime) * 1e-9f;
			result.offset = getOffset();
			result.offsetLen = result.offset.getLength3();
			result.precision = precision;
			result.integral *= DT;
			results.add(result);
			
			if (log.isInfoEnabled())
			{
				StringBuilder sb = new StringBuilder();
				sb.append("Evaluation took ");
				sb.append(result.time);
				sb.append("s. Offset: ");
				sb.append(String.format("%5.5f %5.5f %5.5f", result.offset.x(), result.offset.y(), result.offset.z()));
				sb.append(", offsetLen: ");
				sb.append(String.format("%5.5f", result.offsetLen));
				sb.append("m (");
				sb.append(precision);
				sb.append("m) ");
				sb.append(result.integral);
				sb.append("m - ");
				sb.append(reason);
				log.info(sb.toString());
			}
			
			nextState(EEvent.EVAL_DONE);
		}
		
		
		@Override
		public Enum<? extends Enum<?>> getIdentifier()
		{
			return EStateId.EVAL;
		}
	}
	
	private class InitState implements IRoleState
	{
		@Override
		public void doEntryActions()
		{
			IMoveToSkill move = AMoveSkill.createMoveToSkill();
			move.getMoveCon().setPenaltyAreaAllowedOur(true);
			move.getMoveCon().updateDestination(initPos);
			move.getMoveCon().updateTargetAngle(0);
			move.setDoComplete(true);
			setNewSkill(move);
		}
		
		
		@Override
		public void doUpdate()
		{
		}
		
		
		@Override
		public void doExitActions()
		{
		}
		
		
		@Override
		public void onSkillStarted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public void onSkillCompleted(final ISkill skill, final BotID botID)
		{
			nextState(EEvent.INIT_DONE);
		}
		
		
		@Override
		public Enum<? extends Enum<?>> getIdentifier()
		{
			return EStateId.INIT;
		}
	}
}
