/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Nov 28, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.test;

import java.io.File;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.circle.Circle;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.rectangle.Rectangle;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.ARole;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.pandora.roles.ERole;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.EFeature;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.AMoveSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.IMoveToSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.ISkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills.PolicyTrainerSkill;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.statemachine.IRoleState;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;
import edu.dhbw.mannheim.tigers.sumatra.util.learning.GaussianPolicy;


/**
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class PolicyTrainerRole extends ARole
{
	private static final Logger	log					= Logger.getLogger(PolicyTrainerRole.class.getName());
	
	@Configurable(comment = "reset if bot is outside field with this tolerance")
	private static float				outsideFieldTol	= -200;
	
	private final Rectangle			field;
	private GaussianPolicy			gp						= null;
	private final String				ctrlFile;
	private final String				dataFile;
	
	
	/**
	 * @param ctrlFile
	 * @param dataFile
	 * @param fieldP1
	 * @param fieldP2
	 */
	public PolicyTrainerRole(final String ctrlFile, final String dataFile, final IVector2 fieldP1, final IVector2 fieldP2)
	{
		super(ERole.POLICY_TRAINER);
		this.ctrlFile = ctrlFile;
		this.dataFile = dataFile;
		field = new Rectangle(fieldP1, fieldP2);
		IRoleState posState = new PositioningState();
		IRoleState trainState = new TrainerState();
		IRoleState waitState = new WaitState();
		setInitialState(waitState);
		addTransition(EStateId.POSITIONING, EEvent.READY, trainState);
		addTransition(EStateId.TRAIN, EEvent.OUTSIDE_FIELD, posState);
		addTransition(EStateId.WAIT, EEvent.OUTSIDE_FIELD, posState);
		addTransition(EStateId.WAIT, EEvent.READY, trainState);
		addTransition(EStateId.TRAIN, EEvent.ROLLOUTS_DONE, waitState);
	}
	
	
	private enum EStateId
	{
		POSITIONING,
		TRAIN,
		WAIT
	}
	
	private enum EEvent
	{
		OUTSIDE_FIELD,
		READY,
		ROLLOUTS_DONE
	}
	
	
	@Override
	public void fillNeededFeatures(final List<EFeature> features)
	{
	}
	
	
	private IVector2 getInitPos()
	{
		return field.getMidPoint();
	}
	
	private class PositioningState implements IRoleState
	{
		
		@Override
		public void doEntryActions()
		{
			IMoveToSkill skill = AMoveSkill.createMoveToSkill();
			skill.getMoveCon().updateDestination(getInitPos());
			skill.getMoveCon().updateTargetAngle(0);
			skill.getMoveCon().setPenaltyAreaAllowedOur(true);
			skill.setDoComplete(true);
			setNewSkill(skill);
		}
		
		
		@Override
		public void doUpdate()
		{
			Circle c = new Circle(getInitPos(), 500);
			if (c.isPointInShape(getPos()))
			{
				nextState(EEvent.READY);
			}
		}
		
		
		@Override
		public void doExitActions()
		{
		}
		
		
		@Override
		public void onSkillStarted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public void onSkillCompleted(final ISkill skill, final BotID botID)
		{
			nextState(EEvent.READY);
		}
		
		
		@Override
		public Enum<? extends Enum<?>> getIdentifier()
		{
			return EStateId.POSITIONING;
		}
	}
	
	private class TrainerState implements IRoleState
	{
		PolicyTrainerSkill	skill;
		
		
		@Override
		public void doEntryActions()
		{
			skill = new PolicyTrainerSkill(gp, field);
			setNewSkill(skill);
		}
		
		
		@Override
		public void doUpdate()
		{
			if (!field.isPointInShape(getPos(), outsideFieldTol))
			{
				log.debug("outside of field");
				if (!resetRollout())
				{
					nextState(EEvent.OUTSIDE_FIELD);
				}
			}
		}
		
		
		@Override
		public void doExitActions()
		{
		}
		
		
		@Override
		public void onSkillStarted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public void onSkillCompleted(final ISkill skill, final BotID botID)
		{
			if (!resetRollout())
			{
				this.skill = new PolicyTrainerSkill(gp, field);
				setNewSkill(this.skill);
			}
		}
		
		
		private boolean resetRollout()
		{
			// this is only called if skill actively completed itself
			boolean done = gp.rolloutDone();
			
			if (done)
			{
				nextState(EEvent.ROLLOUTS_DONE);
			} else
			{
				log.debug("Rollouts left: " + (gp.getParams().getNumSamples() - gp.getNumRollouts()));
			}
			return done;
		}
		
		
		@Override
		public Enum<? extends Enum<?>> getIdentifier()
		{
			return EStateId.TRAIN;
		}
	}
	
	private class WaitState implements IRoleState
	{
		File		file;
		long		lastMod	= 0;
		boolean	run		= false;
		
		
		@Override
		public void doEntryActions()
		{
			file = new File(ctrlFile + ".comm");
			log.debug("Waiting for new controller file.");
			
			IMoveToSkill skill = AMoveSkill.createMoveToSkill();
			skill.getMoveCon().updateDestination(getInitPos());
			skill.getMoveCon().updateTargetAngle(0);
			skill.getMoveCon().setPenaltyAreaAllowedOur(true);
			skill.setDoComplete(true);
			setNewSkill(skill);
		}
		
		
		@Override
		public void doUpdate()
		{
			if (file.lastModified() > lastMod)
			{
				lastMod = file.lastModified();
				run = true;
			}
			else if (run)
			{
				run = false;
				gp = GaussianPolicy.createGp(ctrlFile, dataFile);
				nextState(EEvent.OUTSIDE_FIELD);
			}
		}
		
		
		@Override
		public void doExitActions()
		{
		}
		
		
		@Override
		public void onSkillStarted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public void onSkillCompleted(final ISkill skill, final BotID botID)
		{
		}
		
		
		@Override
		public Enum<? extends Enum<?>> getIdentifier()
		{
			return EStateId.WAIT;
		}
	}
}
