/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Oct 30, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver;

import java.util.List;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.IDrawableShape;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * This PathDriver will wrap two drivers. It will use the far one unless the bot is near the destination. Then it uses
 * the close one
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class MixedPathDriver implements IPathDriver
{
	private final IPathDriver	pathDriverFar;
	private final IPathDriver	pathDriverClose;
	
	@Configurable(comment = "Dist [mm] - this is the threshold distance to destination, when the close path driver will be activated")
	private static float			changeToClosePathDriverThreshold	= 100;
	
	
	/**
	 * @param pathDriverFar
	 * @param pathDriverClose
	 */
	public MixedPathDriver(final IPathDriver pathDriverFar, final IPathDriver pathDriverClose)
	{
		this.pathDriverFar = pathDriverFar;
		this.pathDriverClose = pathDriverClose;
	}
	
	
	private IPathDriver getCurrentPathDriver(final TrackedTigerBot bot)
	{
		IVector2 currentPos = bot.getPos();
		IVector3 finalDest = getFinalDestination(bot);
		float dist = GeoMath.distancePP(currentPos, finalDest.getXYVector());
		
		if (dist < (changeToClosePathDriverThreshold))
		{
			return pathDriverClose;
		}
		return pathDriverFar;
	}
	
	
	@Override
	public IVector3 getNextDestination(final TrackedTigerBot bot)
	{
		return getCurrentPathDriver(bot).getNextDestination(bot);
	}
	
	
	@Override
	public IVector3 getNextVelocity(final TrackedTigerBot bot)
	{
		return getCurrentPathDriver(bot).getNextVelocity(bot);
	}
	
	
	@Override
	public IVector3 getFinalDestination(final TrackedTigerBot bot)
	{
		return pathDriverClose.getFinalDestination(bot);
	}
	
	
	@Override
	public void setMovingSpeed(final EMovingSpeed speed)
	{
		pathDriverClose.setMovingSpeed(speed);
		pathDriverFar.setMovingSpeed(speed);
	}
	
	
	@Override
	public float getRemainingTime(final TrackedTigerBot bot)
	{
		return getCurrentPathDriver(bot).getRemainingTime(bot);
	}
	
	
	@Override
	public void decoratePath(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
	{
		getCurrentPathDriver(bot).decoratePath(bot, shapes);
	}
	
	
	@Override
	public void decoratePathDebug(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
	{
		getCurrentPathDriver(bot).decoratePathDebug(bot, shapes);
	}
}
