/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Oct 31, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver;

import java.awt.Color;
import java.util.List;

import Jama.Matrix;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.DrawableBot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.IDrawableShape;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.LinearPolicy;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;
import edu.dhbw.mannheim.tigers.sumatra.util.learning.IPolicyController;


/**
 * This driver will drive straight to its destination, but considers far distances
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class PositionDriver implements IPathDriver
{
	private IVector2						destination			= null;
	private IVector3						nextDestination	= Vector3.ZERO_VECTOR;
	private float							orientation			= 0;
	
	
	@Configurable(comment = "max dist [mm] that destination may differ from bot pos. If greater, dest is modified.", spezis = {
			"", "TIGER_V3" })
	private static float					maxDistanceFast	= 400;
	
	@Configurable(comment = "Max dist [mm] in slow mode", spezis = { "", "TIGER_V3" })
	private static float					maxDistanceSlow	= 200;
	
	@Configurable(comment = "Max dist [mm] in faster mode", spezis = { "", "TIGER_V3" })
	private static float					maxDistanceFaster	= 500;
	
	
	private float							maxDistance			= maxDistanceFast;
	
	private final IPolicyController	velocityPolicy		= new LinearPolicy();
	
	
	@Override
	public IVector3 getNextDestination(final TrackedTigerBot bot)
	{
		if (destination == null)
		{
			return new Vector3(bot.getPos(), bot.getAngle());
		}
		IVector2 dest = destination;
		float dist = GeoMath.distancePP(dest, bot.getPos());
		if (dist > maxDistance)
		{
			dest = GeoMath.stepAlongLine(bot.getPos(), dest, maxDistance);
		}
		nextDestination = new Vector3(dest, orientation);
		return nextDestination;
	}
	
	
	@Override
	public IVector3 getNextVelocity(final TrackedTigerBot bot)
	{
		IVector3 destination = getNextDestination(bot);
		IVector2 error = destination.getXYVector().subtractNew(bot.getPos()).multiply(0.003f);
		float errorW = destination.z() - bot.getAngle();
		
		
		double[] stateArr = new double[] { error.x(), error.y(), errorW, bot.getVel().x(), bot.getVel().y(),
				bot.getaVel() };
		Matrix state = new Matrix(stateArr, 1);
		
		Matrix u = velocityPolicy.getControl(state);
		return new Vector3(u.get(0, 0), u.get(0, 1), u.get(0, 2));
	}
	
	
	@Override
	public IVector3 getFinalDestination(final TrackedTigerBot bot)
	{
		if (destination == null)
		{
			return new Vector3(bot.getPos(), orientation);
		}
		return new Vector3(destination, orientation);
	}
	
	
	@Override
	public void setMovingSpeed(final EMovingSpeed speed)
	{
		switch (speed)
		{
			case FAST:
				maxDistance = maxDistanceFaster;
				break;
			case NORMAL:
				maxDistance = maxDistanceFast;
				break;
			case SLOW:
				maxDistance = maxDistanceSlow;
				break;
			case UNLIMITED:
				maxDistance = Float.MAX_VALUE;
				break;
			default:
				break;
		}
	}
	
	
	@Override
	public float getRemainingTime(final TrackedTigerBot bot)
	{
		// calculate an approximate time
		return 0;
	}
	
	
	/**
	 * @return the destination
	 */
	public IVector2 getDestination()
	{
		return destination;
	}
	
	
	/**
	 * @param destination the destination to set
	 */
	public void setDestination(final IVector2 destination)
	{
		this.destination = destination;
	}
	
	
	/**
	 * @param destination the destination to set
	 */
	public void setDestination(final IVector3 destination)
	{
		this.destination = destination.getXYVector();
		orientation = destination.z();
	}
	
	
	/**
	 * @return the orientation
	 */
	public float getOrientation()
	{
		return orientation;
	}
	
	
	/**
	 * @param orientation the orientation to set
	 */
	public void setOrientation(final float orientation)
	{
		this.orientation = orientation;
	}
	
	
	@Override
	public void decoratePath(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
	{
		// draw current destination
		shapes.add(new DrawableBot(nextDestination.getXYVector(), nextDestination.z(), Color.red));
	}
	
	
	@Override
	public void decoratePathDebug(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
	{
		// draw current destination
		shapes.add(new DrawableBot(nextDestination.getXYVector(), nextDestination.z(), Color.red));
	}
}
