/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Nov 30, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots;

import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import com.sleepycat.persist.model.Persistent;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.trajectory.SplinePair3D;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.basestation.IBaseStation;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.communication.ENetworkState;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.communication.ReliableCmdManager;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.communication.Statistics;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.CommandFactory;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerDribble;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerChargeAuto;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2.EKickerMode;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorMoveV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerCtrlSetPIDParams;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSkillPositioningCommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSystemMatchCtrl;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSystemMatchFeedback;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSystemMatchPosVel;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.EKickDevice;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * New Bot 2015 with less/no knowledge about itself in Sumatra
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
@Persistent
public class TigerBotV3 extends ABot
{
	private static final Logger						log						= Logger.getLogger(TigerBotV3.class.getName());
	
	/**  */
	public static final float							BAT_MIN					= 10.5f;
	/**  */
	public static final float							BAT_MAX					= 12.6f;
	
	private float											battery					= 0;
	private float											kickerLevel				= 0;
	
	private transient final IBaseStation			baseStation;
	
	private final Statistics							txStats					= new Statistics();
	private final Statistics							rxStats					= new Statistics();
	
	private transient final ReliableCmdManager	reliableCmdManager	= new ReliableCmdManager(this);
	
	
	@Configurable(comment = "Desired frequency [Hz] of feedback commands from bot if idle")
	private static int									feedbackFreqIdle		= 5;
	
	@Configurable(comment = "Desired frequency [Hz] of feedback commands from bot if active")
	private static int									feedbackFreqActive	= 120;
	
	@Configurable(comment = "How many pakets should be skipped in idle mode?")
	private static int									skipPacketsOnIdle		= 50;
	
	@Configurable(comment = "Bot names, ordered by id")
	private static String[]								botNames					= { "Gandalf", "Achilles", "Eichbaum", "Tigger",
																							"Optimus Prime", "Odysseus", "Leopard 3", "Q",
																							"Reality Checkpoint", "This Bot", "Yoda",
																							"Hercules" };
	
	@Configurable(comment = "Disable controller on reset")
	private static boolean								disableControllers	= true;
	
	
	private transient TigerSystemMatchFeedback	latestFeedbackCmd		= new TigerSystemMatchFeedback();
	
	private MatchCmdMemory								matchCmdMemory			= new MatchCmdMemory();
	
	@Persistent
	private static class MatchCmdMemory
	{
		private boolean	controllerOn	= false;
		private boolean	cheer				= false;
		private int			kickSeq			= 0;
		private float		kickDuration	= 0;
		private int			kickMode			= EKickerMode.DISARM.getId();
		private int			kickDevice		= EKickDevice.STRAIGHT.getValue();
		private long		tUpdated			= 0;
		private IVector3	vel				= null;
		private IVector3	pos				= null;
		private float		dribbleSpeed	= 0;
		private boolean	chargeAuto		= false;
		
		
		@Override
		public String toString()
		{
			StringBuilder builder = new StringBuilder();
			builder.append("MatchCmdMemory [controllerOn=");
			builder.append(controllerOn);
			builder.append(", cheer=");
			builder.append(cheer);
			builder.append(", kickSeq=");
			builder.append(kickSeq);
			builder.append(", kickDuration=");
			builder.append(kickDuration);
			builder.append(", kickMode=");
			builder.append(kickMode);
			builder.append(", kickDevice=");
			builder.append(kickDevice);
			builder.append(", tUpdated=");
			builder.append(tUpdated);
			builder.append(", vel=");
			builder.append(vel);
			builder.append(", pos=");
			builder.append(pos);
			builder.append(", dribbleSpeed=");
			builder.append(dribbleSpeed);
			builder.append(", chargeAuto=");
			builder.append(chargeAuto);
			builder.append("]");
			return builder.toString();
		}
	}
	
	
	@SuppressWarnings("unused")
	private TigerBotV3()
	{
		baseStation = null;
	}
	
	
	/**
	 * @param botId
	 * @param baseStation
	 */
	public TigerBotV3(final BotID botId, final IBaseStation baseStation)
	{
		super(EBotType.TIGER_V3, botId, 0, -1);
		this.baseStation = baseStation;
		setName(botNames[botId.getNumber()]);
		log.debug("New TigerBot V3 with ID" + botId);
	}
	
	
	@Override
	public void setDefaultKickerMaxCap()
	{
	}
	
	
	@Override
	protected Map<EFeature, EFeatureState> getDefaultFeatureStates()
	{
		Map<EFeature, EFeatureState> result = EFeature.createFeatureList();
		result.put(EFeature.DRIBBLER, EFeatureState.WORKING);
		result.put(EFeature.CHIP_KICKER, EFeatureState.WORKING);
		result.put(EFeature.STRAIGHT_KICKER, EFeatureState.WORKING);
		result.put(EFeature.MOVE, EFeatureState.WORKING);
		result.put(EFeature.BARRIER, EFeatureState.WORKING);
		return result;
	}
	
	
	@Override
	public void execute(final ACommand cmd)
	{
		boolean handled = handleMatchCmd(cmd);
		if (handled)
		{
			matchCmdMemory.tUpdated = SumatraClock.nanoTime();
		} else
		{
			switch (cmd.getType())
			{
				case CMD_CTRL_SET_PID_PARAMS:
					TigerCtrlSetPIDParams pids = (TigerCtrlSetPIDParams) cmd;
					log.debug("Set PIDs for " + pids.getParamType() + " on " + getBotID() + " to " + pids.getParams());
					break;
				default:
			}
			
			// command not considered in match cmd, send it raw
			executeCommand(cmd);
		}
	}
	
	
	/**
	 * Execute a cmd directly, no match cmd handling here!
	 * 
	 * @param cmd
	 */
	private void executeCommand(final ACommand cmd)
	{
		reliableCmdManager.outgoingCommand(cmd);
		baseStation.enqueueCommand(getBotID(), cmd);
		txStats.packets++;
		txStats.payload += CommandFactory.getInstance().getLength(cmd, false);
	}
	
	
	private boolean handleMatchCmd(final ACommand cmd)
	{
		switch (cmd.getType())
		{
			case CMD_CTRL_RESET:
				matchCmdMemory.controllerOn = !disableControllers;
				matchCmdMemory.vel = null;
				matchCmdMemory.pos = null;
				reliableCmdManager.clearAllPendingCmds();
				matchCmdMemory.dribbleSpeed = 0;
				if (matchCmdMemory.kickMode != EKickerMode.DISARM.getId())
				{
					matchCmdMemory.kickMode = EKickerMode.DISARM.getId();
					matchCmdMemory.kickSeq++;
				}
				return true;
			case CMD_KICKER_KICKV2:
				TigerKickerKickV2 kick = (TigerKickerKickV2) cmd;
				matchCmdMemory.kickSeq++;
				matchCmdMemory.kickDuration = kick.getFiringDuration();
				matchCmdMemory.kickDevice = kick.getDevice();
				matchCmdMemory.kickMode = kick.getMode();
				return true;
			case CMD_MOTOR_DRIBBLE:
				TigerDribble dribble = (TigerDribble) cmd;
				matchCmdMemory.dribbleSpeed = dribble.getSpeed();
				return true;
			case CMD_MOTOR_MOVE_V2:
				TigerMotorMoveV2 motor = (TigerMotorMoveV2) cmd;
				matchCmdMemory.vel = new Vector3(motor.getXY(), motor.getW());
				matchCmdMemory.pos = null;
				matchCmdMemory.controllerOn = true;
				return true;
			case CMD_SKILL_POSITIONING:
				TigerSkillPositioningCommand pos = (TigerSkillPositioningCommand) cmd;
				matchCmdMemory.pos = new Vector3(pos.getDestination().multiplyNew(1e-3f), pos.getOrientation());
				matchCmdMemory.vel = null;
				matchCmdMemory.controllerOn = true;
				return true;
			case CMD_SYSTEM_MATCH_POS_VEL:
				TigerSystemMatchPosVel posVel = (TigerSystemMatchPosVel) cmd;
				matchCmdMemory.pos = new Vector3(posVel.getDestination().multiplyNew(1e-3f), posVel.getOrientation());
				matchCmdMemory.vel = posVel.getVelocity();
				matchCmdMemory.controllerOn = true;
				return true;
			case CMD_KICKER_CHARGE_AUTO:
				TigerKickerChargeAuto charge = (TigerKickerChargeAuto) cmd;
				if (charge.getMax() > 0)
				{
					if (!matchCmdMemory.chargeAuto)
					{
						matchCmdMemory.chargeAuto = true;
						matchCmdMemory.kickSeq++;
					}
				} else if (matchCmdMemory.chargeAuto)
				{
					matchCmdMemory.chargeAuto = false;
					matchCmdMemory.kickSeq++;
				}
				return true;
			default:
				break;
		}
		return false;
	}
	
	
	/**
	 * Translate given cmds to a match cmd and send it.
	 * This will be called regularly by SkillExecutor
	 * 
	 * @param cmds
	 */
	public void executeMatchCmd(final List<ACommand> cmds)
	{
		for (ACommand cmd : cmds)
		{
			execute(cmd);
		}
		
		final int feedbackFreq;
		if ((SumatraClock.nanoTime() - matchCmdMemory.tUpdated) < 1e8)
		{
			feedbackFreq = feedbackFreqActive;
		} else
		{
			feedbackFreq = feedbackFreqIdle;
		}
		
		
		TigerSystemMatchCtrl match = new TigerSystemMatchCtrl();
		
		if (matchCmdMemory.vel != null)
		{
			match.setVelocity(matchCmdMemory.vel.getXYVector(), matchCmdMemory.vel.z());
		} else
		{
			match.resetVelocity();
		}
		if (matchCmdMemory.pos != null)
		{
			match.setPosition(matchCmdMemory.pos.getXYVector(), matchCmdMemory.pos.z());
		} else
		{
			match.resetPosition();
		}
		
		match.setKickerAutocharge(matchCmdMemory.chargeAuto, matchCmdMemory.kickSeq);
		match.setKick(matchCmdMemory.kickDuration, matchCmdMemory.kickDevice, matchCmdMemory.kickMode,
				matchCmdMemory.kickSeq);
		match.setController(matchCmdMemory.controllerOn);
		match.setCheering(matchCmdMemory.cheer);
		match.setFeedbackFreq(feedbackFreq);
		match.setDribblerSpeed(matchCmdMemory.dribbleSpeed);
		executeCommand(match);
	}
	
	
	@Override
	public void start()
	{
	}
	
	
	@Override
	public void stop()
	{
	}
	
	
	@Override
	public float getBatteryLevel()
	{
		return battery;
	}
	
	
	@Override
	public float getBatteryLevelMax()
	{
		return BAT_MAX;
	}
	
	
	@Override
	public float getBatteryLevelMin()
	{
		return BAT_MIN;
	}
	
	
	@Override
	public float getKickerLevel()
	{
		return kickerLevel;
	}
	
	
	@Override
	public float getKickerLevelMax()
	{
		return 180f;
	}
	
	
	@Override
	public ENetworkState getNetworkState()
	{
		return ENetworkState.ONLINE;
	}
	
	
	@Override
	public void newSpline(final SplinePair3D spline)
	{
	}
	
	
	private void onNewFeedbackCmd(final TigerSystemMatchFeedback cmd)
	{
		for (EFeature f : EFeature.values())
		{
			getBotFeatures().put(f, cmd.isFeatureWorking(f) ? EFeatureState.WORKING : EFeatureState.KAPUT);
		}
	}
	
	
	@Override
	public void onIncommingBotCommand(final BotID id, final ACommand cmd)
	{
		if (!id.equals(getBotID()))
		{
			return;
		}
		
		reliableCmdManager.incommingCommand(cmd);
		
		switch (cmd.getType())
		{
			case CMD_SYSTEM_MATCH_FEEDBACK:
				latestFeedbackCmd = (TigerSystemMatchFeedback) cmd;
				battery = latestFeedbackCmd.getBatteryLevel();
				kickerLevel = latestFeedbackCmd.getKickerLevel();
				onNewFeedbackCmd(latestFeedbackCmd);
				break;
			default:
				break;
		}
		
		rxStats.packets++;
		rxStats.payload += CommandFactory.getInstance().getLength(cmd, false);
	}
	
	
	/**
	 * @return
	 */
	@Override
	public Statistics getRxStats()
	{
		return new Statistics(rxStats);
	}
	
	
	/**
	 * @return
	 */
	@Override
	public Statistics getTxStats()
	{
		return new Statistics(txStats);
	}
	
	
	/**
	 * @return the latestFeedbackCmd
	 */
	public TigerSystemMatchFeedback getLatestFeedbackCmd()
	{
		return latestFeedbackCmd;
	}
	
	
	/**
	 * @return
	 */
	@Override
	public int getHardwareId()
	{
		return latestFeedbackCmd.getHardwareId();
	}
	
	
	/**
	 * Create a {@link TrackedTigerBot} from known feedback
	 * 
	 * @return
	 */
	public TrackedTigerBot createTrackedBot()
	{
		IVector2 pos = latestFeedbackCmd.getPosition().multiply(1000);
		IVector2 vel = latestFeedbackCmd.getVelocity();
		IVector2 acc = latestFeedbackCmd.getAcceleration();
		int height = 135;
		float angle = latestFeedbackCmd.getOrientation();
		float aVel = latestFeedbackCmd.getAngularVelocity();
		float aAcc = latestFeedbackCmd.getAngularAcceleration();
		return new TrackedTigerBot(getBotId(), pos, vel, acc, height, angle, aVel, aAcc, 1, this, getBotId()
				.getTeamColor());
	}
	
	
	@Override
	public boolean isAvailableToAi()
	{
		if (super.isAvailableToAi())
		{
			return latestFeedbackCmd.isFeatureWorking(EFeature.MOVE);
		}
		return false;
	}
	
	
	/**
	 * Sets cheering flag
	 * 
	 * @param cheering
	 */
	public void setCheering(final boolean cheering)
	{
		matchCmdMemory.cheer = cheering;
	}
}
