/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Mar 12, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * MarkG <Mark.Geiger@dlr.de>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.DynamicPosition;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.EMovingSpeed;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.AroundBallDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.DoNothingDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.ESkillDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.ISkillDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.KickBallDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.skills.PushBallDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.EFeature;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.EFeatureState;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2.EKickerMode;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.ESkillName;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.ChipParams;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.EKickDevice;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.TigerDevices;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * Perform straight (and chip) kicks with static and moving balls
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 *         MarkG <Mark.Geiger@dlr.de>
 */
public class KickSkill extends AMoveSkill
{
	private static final Logger	log							= Logger.getLogger(KickSkill.class
																					.getName());
	
	// ###### parameters ##########
	// ###### delay #######
	@Configurable(comment = "Delay [s] to wait before sending kick command, starting when kicking is possible")
	private static float				workingKickDelay			= 0.0f;
	@Configurable(comment = "Delay [s] to wait before sending kick command, starting when kicking is possible")
	private static float				brokenKickDelay			= 0.7f;
	private float						kickDelay					= workingKickDelay;
	
	
	@Configurable(comment = "Angle tol [rad] when a kick cmd is sent.")
	private static float				aimingTolerance			= 0.05f;
	
	
	@Configurable(comment = "Which type of command should be send to the bots? POS or VEL")
	private static ECommandType	cmdMode						= ECommandType.POS;
	
	
	// ######### from construction #########
	private DynamicPosition			receiver;
	private EKickMode					kickMode;
	
	// ########## state variables ############
	
	protected long						kickTimeStart				= 0;
	private int							duration						= 0;
	private EMoveMode					moveMode						= EMoveMode.NORMAL;
	private EKickDevice				device						= EKickDevice.STRAIGHT;
	private ISkillDriver				currentDriver				= new DoNothingDriver();
	
	private static final int		BALL_VEL_BUFFER_SIZE		= 20;
	private static final int		ANGLE_DIFF_BUFFER_SIZE	= 15;
	private List<Float>				ballVelBuffer				= new ArrayList<>(BALL_VEL_BUFFER_SIZE);
	private List<Float>				angleDiffBuffer			= new ArrayList<>(ANGLE_DIFF_BUFFER_SIZE);
	
	// ########## Pushing ####################
	private long						timeLastMoved				= System.nanoTime();
	private IVector2					lastPos						= null;
	
	/**
	 */
	public enum EKickMode
	{
		/**  */
		PASS,
		/**  */
		MAX,
		/**  */
		POINT,
		/**  */
		KICKOFF,
		/**  */
		FIXED_DURATION,
	}
	
	/**
	 */
	public enum EMoveMode
	{
		/**  */
		NORMAL,
		/**  */
		CHILL
	}
	
	
	/**
	 * If receiver is a vector, kick in this direction with max force,
	 * if receiver is a botId, pass to this bot with appropriate force
	 * 
	 * @param receiver
	 * @param kickMode
	 */
	public KickSkill(final DynamicPosition receiver, final EKickMode kickMode)
	{
		this(ESkillName.KICK, receiver, kickMode, EMoveMode.NORMAL);
	}
	
	
	/**
	 * If receiver is a vector, kick in this direction with max force,
	 * if receiver is a botId, pass to this bot with appropriate force
	 * 
	 * @param receiver
	 * @param kickMode
	 * @param moveMode
	 */
	public KickSkill(final DynamicPosition receiver, final EKickMode kickMode, final EMoveMode moveMode)
	{
		this(ESkillName.KICK, receiver, kickMode, moveMode);
	}
	
	
	/**
	 * Kick with fixed duration
	 * 
	 * @param receiver
	 * @param moveMode
	 * @param duration
	 */
	public KickSkill(final DynamicPosition receiver, final EMoveMode moveMode, final int duration)
	{
		this(receiver, EKickMode.FIXED_DURATION, moveMode);
		this.duration = duration;
		
	}
	
	
	/**
	 * If receiver is a vector, kick in this direction with max force,
	 * if receiver is a botId, pass to this bot with appropriate force
	 * 
	 * @param skillName
	 * @param receiver
	 */
	protected KickSkill(final ESkillName skillName, final DynamicPosition receiver, final EKickMode kickMode,
			final EMoveMode moveMode)
	{
		super(skillName);
		this.moveMode = moveMode;
		this.receiver = receiver;
		this.kickMode = kickMode;
		setCommandType(cmdMode);
		getMoveCon().setDriveFast(true);
	}
	
	
	private void updateBallVelBuffer()
	{
		if (ballVelBuffer.size() >= (BALL_VEL_BUFFER_SIZE - 1))
		{
			ballVelBuffer.remove(0);
		}
		ballVelBuffer.add(getWorldFrame().getBall().getVel().getLength2());
	}
	
	
	private void updatePushingCheck()
	{
		if (lastPos == null)
		{
			lastPos = getPos();
		}
		float dist2Ball = GeoMath.distancePP(getPos(), getWorldFrame().getBall().getPos());
		if (dist2Ball > 400)
		{
			// we may just wait to receive the ball. pushing only makes sense if all is near by
			return;
		}
		if (GeoMath.distancePP(lastPos, getPos()) > 50)
		{
			timeLastMoved = System.nanoTime();
			lastPos = getPos();
		}
		if (((System.nanoTime() - timeLastMoved) > 1e9) && (currentDriver.getType() != ESkillDriver.PUSH_BALL))
		{
			currentDriver = new PushBallDriver(receiver);
		}
	}
	
	
	private void updateReceiverTarget()
	{
		receiver.update(getWorldFrame());
	}
	
	
	@Override
	protected final void update(final List<ACommand> cmds)
	{
		updateReceiverTarget();
		updateDirection();
		updateBallVelBuffer();
		updateOrientationDiffBuffer();
		updatePushingCheck();
		updateDriver();
		
		generateKickCmd(cmds);
	}
	
	
	private void updateDriver()
	{
		if (currentDriver.isDone() || (currentDriver.getType() == ESkillDriver.DO_NOTHING))
		{
			switch (currentDriver.getType())
			{
				case AROUND_BALL:
					currentDriver = new KickBallDriver(receiver);
					break;
				case KICK_BALL:
				case DO_NOTHING:
				case PUSH_BALL:
					currentDriver = new AroundBallDriver(receiver);
					break;
			}
		}
		setPathDriver(currentDriver);
		currentDriver.update(getTBot(), getWorldFrame());
	}
	
	
	@Override
	public void doCalcEntryActions(final List<ACommand> cmds)
	{
		if (getBot().getBotFeatures().get(EFeature.BARRIER) == EFeatureState.KAPUT)
		{
			kickDelay = brokenKickDelay;
		} else
		{
			kickDelay = workingKickDelay;
		}
		
		setCommandType(cmdMode);
		if (moveMode == EMoveMode.CHILL)
		{
			getMoveCon().setSpeed(EMovingSpeed.SLOW, 1.5f);
		} else
		{
			getMoveCon().setDriveFast(true);
		}
	}
	
	
	@Override
	protected void doCalcExitActions(final List<ACommand> cmds)
	{
		getDevices().allOff(cmds);
	}
	
	
	private void generateKickCmd(final List<ACommand> cmds)
	{
		float kickLength = receiver.subtractNew(getWorldFrame().getBall().getPos()).getLength2();
		EKickDevice usedDevice = device;
		EKickerMode mode = EKickerMode.ARM;
		float duration = TigerDevices.getStraightKickMaxDuration();
		int dribble = 0;
		
		if ((getBot().getBotFeatures().get(EFeature.BARRIER) == EFeatureState.KAPUT))
		{
			mode = EKickerMode.FORCE;
		}
		
		if ((device == EKickDevice.STRAIGHT)
				&& (getBot().getBotFeatures().get(EFeature.STRAIGHT_KICKER) != EFeatureState.WORKING))
		{
			// kicker is broken, lets chip with 1/2 length :)
			kickLength /= 2;
			usedDevice = EKickDevice.CHIP;
		} else if ((device == EKickDevice.CHIP)
				&& (getBot().getBotFeatures().get(EFeature.CHIP_KICKER) != EFeatureState.WORKING))
		{
			usedDevice = EKickDevice.STRAIGHT;
		} else if ((getBot().getBotFeatures().get(EFeature.STRAIGHT_KICKER) != EFeatureState.WORKING)
				&& (getBot().getBotFeatures().get(EFeature.CHIP_KICKER) != EFeatureState.WORKING))
		{
			log.warn("Kicker and Chipper of bot " + getBot().getBotID() + " is configured broken!");
			complete();
		}
		
		switch (usedDevice)
		{
			case CHIP:
			{
				ChipParams chipValues = calcChipParams(kickLength);
				duration = chipValues.getDuration();
				dribble = chipValues.getDribbleSpeed();
				if ((moveMode == EMoveMode.CHILL) && (chipValues.getDribbleSpeed() > 0))
				{
					mode = EKickerMode.DRIBBLER;
				}
				break;
			}
			case STRAIGHT:
			{
				duration = calcStraightDuration(kickLength);
				break;
			}
		}
		
		if (currentDriver.getType() == ESkillDriver.KICK_BALL)
		{
			// this is only used if barrier is broken!
			if (kickTimeStart == 0)
			{
				// start time for kick
				kickTimeStart = SumatraClock.nanoTime();
			}
		} else
		{
			kickTimeStart = 0;
		}
		
		if (readyForKick())
		{
			getDevices().kickGeneral(cmds, mode, usedDevice, duration, dribble);
		} else
		{
			getDevices().dribble(cmds, dribble);
			getDevices().disarm(cmds);
		}
	}
	
	
	private void updateOrientationDiffBuffer()
	{
		if (angleDiffBuffer.size() >= (ANGLE_DIFF_BUFFER_SIZE - 1))
		{
			angleDiffBuffer.remove(0);
		}
		float orientationDiff = Math.abs(AngleMath.getShortestRotation(
				receiver.subtractNew(getWorldFrame().getBall().getPos()).getAngle(), getAngle()));
		angleDiffBuffer.add(orientationDiff);
	}
	
	
	private float getAvgOrientationDiff()
	{
		return (float) angleDiffBuffer.stream().mapToDouble(a -> a).average().getAsDouble();
	}
	
	
	private float getAvgBallSpeed()
	{
		return (float) ballVelBuffer.stream().mapToDouble(a -> a).average().getAsDouble();
	}
	
	
	private boolean readyForKick()
	{
		boolean delayReached = (kickTimeStart != 0)
				&& (TimeUnit.NANOSECONDS.toMillis(SumatraClock.nanoTime() - kickTimeStart) >= (kickDelay * 1000));
		
		final float orientDiffAvg = getAvgOrientationDiff();
		boolean botOrientationSatisfied = orientDiffAvg < aimingTolerance;
		
		return (currentDriver.getType() != ESkillDriver.PUSH_BALL) && delayReached && botOrientationSatisfied;
	}
	
	
	private void updateDirection()
	{
		receiver.update(getWorldFrame());
	}
	
	
	protected int calcStraightDuration(final float length)
	{
		int duration;
		switch (kickMode)
		{
			case MAX:
				duration = TigerDevices.getStraightKickMaxDuration();
				break;
			case PASS:
				final float passEndVel = getAvgBallSpeed();
				duration = (int) TigerDevices.calcStraightDuration(length, passEndVel);
				break;
			case POINT:
				duration = (int) TigerDevices.calcStraightDuration(length, 0);
				break;
			case KICKOFF:
				// duration = (int) TigerDevices.calcStraightDuration(length, passEndVel);
				duration = 1900;
				break;
			case FIXED_DURATION:
				duration = this.duration;
				break;
			default:
				throw new IllegalStateException();
		}
		
		return duration;
	}
	
	
	protected ChipParams calcChipParams(final float kickLength)
	{
		return TigerDevices.calcChipFastParams(kickLength);
	}
	
	
	/**
	 * @return the kickDelay
	 */
	public final float getKickDelay()
	{
		return kickDelay;
	}
	
	
	/**
	 * @param kickDelay the kickDelay to set
	 */
	public final void setKickDelay(final float kickDelay)
	{
		this.kickDelay = kickDelay;
	}
	
	
	/**
	 * @return the device
	 */
	public final EKickDevice getDevice()
	{
		return device;
	}
	
	
	/**
	 * @param device the device to set
	 */
	public final void setDevice(final EKickDevice device)
	{
		this.device = device;
	}
	
	
	/**
	 * Update the receiver target
	 * 
	 * @param recv
	 */
	public final void setReceiver(final DynamicPosition recv)
	{
		receiver = recv;
		updateDirection();
	}
	
	
	/**
	 * @return the kickMode
	 */
	public final EKickMode getKickMode()
	{
		return kickMode;
	}
	
	
	/**
	 * @param kickMode the kickMode to set
	 */
	public final void setKickMode(final EKickMode kickMode)
	{
		this.kickMode = kickMode;
	}
	
	
	/**
	 * @return the receiver
	 */
	public final DynamicPosition getReceiver()
	{
		return receiver;
	}
}
