/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Nov 16, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills;

import java.util.LinkedList;
import java.util.List;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.IDrawableShape;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.AVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.EMovingSpeed;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.IPathDriver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.ESkillName;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * Drive a letter by given path points
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class LetterSkill extends AMoveSkill
{
	private final List<IVector2>	pathPoints;
	private List<IVector2>			morePathPoints;
	
	@Configurable(comment = "If nearer than this dist [mm] to next point, go on to next point.")
	private static float				distTolerance	= 75;
	
	@Configurable
	private static int				resolution		= 100;
	
	
	/**
	 * @param pathPoints
	 */
	public LetterSkill(final List<IVector2> pathPoints)
	{
		super(ESkillName.LETTER);
		this.pathPoints = pathPoints;
		morePathPoints = new LinkedList<IVector2>();
		IVector3 initialDestination = new Vector3(pathPoints.get(0), 0);
		// use our own pathdriver, you can also try existing Implementations, like PathPointDriver
		setPathDriver(new PathDriver(initialDestination));
	}
	
	
	@Override
	protected void doCalcEntryActions(final List<ACommand> cmds)
	{
		
		for (int i = 0; i < (pathPoints.size() - 1); i++)
		{
			double distance = GeoMath.distancePP(pathPoints.get(i), pathPoints.get(i + 1));
			int n = (int) (distance / resolution);
			
			IVector2 direction = pathPoints.get(i + 1).subtractNew(pathPoints.get(i)).normalizeNew()
					.multiplyNew(resolution);
			if (!morePathPoints.isEmpty() && !pathPoints.get(i).equals(morePathPoints.get(morePathPoints.size() - 1)))
			{
				morePathPoints.add(pathPoints.get(i));
			}
			
			for (int j = 1; j <= n; j++)
			{
				morePathPoints.add(pathPoints.get(i).addNew(direction.multiplyNew(j)));
			}
		}
	}
	
	
	@Override
	protected void update(final List<ACommand> cmds)
	{
		// executed continuously
	}
	
	
	@Override
	protected void doCalcExitActions(final List<ACommand> cmds)
	{
		// executed once at the end
	}
	
	private class PathDriver implements IPathDriver
	{
		IVector3	currentDestination;
		int		currentDestinationIndex;
		
		
		public PathDriver(final IVector3 initialDestination)
		{
			currentDestination = initialDestination;
			currentDestinationIndex = 0;
		}
		
		
		@Override
		public IVector3 getNextDestination(final TrackedTigerBot bot)
		{
			if ((GeoMath.distancePP(currentDestination.getXYVector(), bot.getPos()) < distTolerance)
					&& (currentDestinationIndex < (morePathPoints.size() - 1)))
			{
				currentDestinationIndex++;
				currentDestination = new Vector3(morePathPoints.get(currentDestinationIndex), morePathPoints
						.get(currentDestinationIndex).subtractNew(morePathPoints.get(currentDestinationIndex - 1)).getAngle());
			}
			
			return currentDestination;
		}
		
		
		@Override
		public IVector3 getNextVelocity(final TrackedTigerBot bot)
		{
			// TODO return a Vector: [xd,yd,rotationd]
			return AVector3.ZERO_VECTOR;
		}
		
		
		@Override
		public IVector3 getFinalDestination(final TrackedTigerBot bot)
		{
			// TODO return a Vector: [x,y,rotation]
			return new Vector3(pathPoints.get(pathPoints.size() - 1), 0);
		}
		
		
		@Override
		public void setMovingSpeed(final EMovingSpeed speed)
		{
			// not relevant for you
		}
		
		
		@Override
		public float getRemainingTime(final TrackedTigerBot bot)
		{
			// you do not need to implement this
			return 0;
		}
		
		
		@Override
		public void decoratePath(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
		{
			// add some shapes to the list and they will be drawn if you enable the pathplanning layers
		}
		
		
		@Override
		public void decoratePathDebug(final TrackedTigerBot bot, final List<IDrawableShape> shapes)
		{
			
		}
		
	}
}
