/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Mar 9, 2014
 * Author(s): Mark Geiger <Mark.Geiger@dlr.de>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills;

import java.util.List;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AiMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2.EKickerMode;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorMoveV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSkillPositioningCommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.ESkillName;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.EKickDevice;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.TigerDevices;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * Super awesome special moving kick.
 * 
 * @author Mark Geiger <Mark.Geiger@dlr.de>
 */
public class MovingKickSkill extends ASkill
{
	
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	
	private IVector2			bestShootTarget			= null;
	
	private IVector2			initialMovePosition		= null;
	
	private IVector2			initStepPos					= null;
	
	@Configurable
	private static float		forwardMoveSpeed			= 0.65f;
	
	@Configurable
	private static float		curveMoveSpeed				= 0.4f;
	
	@Configurable
	private static float		turnAngleSpeed				= 1.2f;
	
	@Configurable
	private static float		finalSpeedBeforeShoot	= 0.8f;
	
	@Configurable
	private static float		safetyDistance				= 270f;
	
	@Configurable
	private static int		dribbleSpeed				= 0;
	
	private EMovementState	movementState				= EMovementState.POSITIONSKILL;
	
	private enum EMovementState
	{
		POSITIONSKILL,
		FORWARD,
		CURVE,
		SHOOT
	}
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	
	
	/**
	 * This Skill moves on a smooth Curve and drags/shoots the ball into the enemy goal.
	 * 
	 * @param target
	 */
	public MovingKickSkill(final IVector2 target)
	{
		this(ESkillName.MOVING_KICK, true);
		bestShootTarget = target;
	}
	
	
	protected MovingKickSkill(final ESkillName skillName, final boolean kickwhenFinished)
	{
		super(skillName);
	}
	
	
	@Override
	protected void doCalcActions(final List<ACommand> cmds)
	{
		TigerMotorMoveV2 move = null;
		IVector2 localVel = null;
		IVector2 direction = null;
		switch (movementState)
		{
			case POSITIONSKILL:
				float orient = getWorldFrame().getBall().getPos().subtractNew(initialMovePosition).getAngle();
				Vector2 dest = new Vector2(initialMovePosition);
				
				if (getWorldFrame().isInverted())
				{
					dest.multiply(-1);
					orient = AngleMath.normalizeAngle(orient + AngleMath.PI);
				}
				cmds.add(new TigerSkillPositioningCommand(dest, orient));
				if (GeoMath.distancePP(getPos(), initialMovePosition) < 30)
				{
					initStepPos = getPos();
					movementState = EMovementState.FORWARD;
				}
				break;
			
			case FORWARD:
				getDevices().dribble(cmds, dribbleSpeed);
				move = new TigerMotorMoveV2();
				direction = getWorldFrame().getBall().getPos().subtractNew(initStepPos).normalizeNew();
				localVel = AiMath.convertGlobalBotVector2Local(direction, getAngle()).normalizeNew()
						.multiplyNew(forwardMoveSpeed);
				
				move.setX(localVel.x());
				move.setY(localVel.y());
				move.setW(AngleMath.normalizeAngle(getWorldFrame().getBall().getPos().getAngle() - getAngle()));
				
				if (GeoMath.distancePP(initStepPos, getPos()) > 500)
				{
					movementState = EMovementState.CURVE;
					initStepPos = getPos();
				}
				cmds.add(move);
				break;
			
			case CURVE:
				getDevices().dribble(cmds, dribbleSpeed);
				move = new TigerMotorMoveV2();
				
				float botToTarget = bestShootTarget.subtractNew(getPos()).getAngle();
				float botOrient = getAngle();
				float angleDif = botToTarget - botOrient;
				if (angleDif > 0)
				{
					angleDif = turnAngleSpeed;
				} else if (angleDif < 0)
				{
					angleDif = -turnAngleSpeed;
				} else
				{
					angleDif = 0;
				}
				
				direction = new Vector2(getAngle());
				localVel = AiMath.convertGlobalBotVector2Local(direction, getAngle()).normalizeNew()
						.multiplyNew(curveMoveSpeed);
				
				move.setY(localVel.y());
				move.setX(localVel.x());
				
				move.setW(AngleMath.normalizeAngle(angleDif));
				if (Math.abs(botToTarget - botOrient) < 0.05)
				{
					movementState = EMovementState.SHOOT;
					initStepPos = getPos();
				}
				
				cmds.add(move);
				break;
			
			case SHOOT:
				move = new TigerMotorMoveV2();
				direction = bestShootTarget.subtractNew(initStepPos).normalizeNew();
				localVel = AiMath.convertGlobalBotVector2Local(direction, getAngle()).normalizeNew()
						.multiplyNew(finalSpeedBeforeShoot);
				
				move.setX(localVel.x());
				move.setY(localVel.y());
				move.setW(0);
				
				if (GeoMath.distancePP(initStepPos, getPos()) > 50)
				{
					getDevices().kickGeneral(cmds, EKickerMode.FORCE, EKickDevice.STRAIGHT,
							TigerDevices.getStraightKickMaxDuration(), 0);
					complete();
				}
				cmds.add(move);
				break;
		}
	}
	
	
	@Override
	protected void doCalcEntryActions(final List<ACommand> cmds)
	{
		initialMovePosition = determineInitialMovePosition();
	}
	
	
	private IVector2 determineInitialMovePosition()
	{
		// can be null, if there are no enemies
		IVector2 enemyBotPos = AiMath.getNearestBot(getWorldFrame().foeBots, getWorldFrame().getBall().getPos()).getPos();
		IVector2 targetToBall = getWorldFrame().getBall().getPos().subtractNew(bestShootTarget);
		IVector2 enemyBotToBall = getWorldFrame().getBall().getPos().subtractNew(enemyBotPos);
		
		IVector2 supportPoint1 = getWorldFrame().getBall().getPos().addNew(
				enemyBotToBall.getNormalVector().multiplyNew(safetyDistance));
		IVector2 supportPoint2 = getWorldFrame().getBall().getPos().addNew(
				enemyBotToBall.getNormalVector().multiplyNew(-safetyDistance));
		
		IVector2 support1ToBall = getWorldFrame().getBall().getPos().subtractNew(supportPoint1);
		IVector2 support2ToBall = getWorldFrame().getBall().getPos().subtractNew(supportPoint2);
		
		// supportToBalls mit targetToBall vergleichen.
		float orientS1 = targetToBall.multiplyNew(-1).getAngle() - support1ToBall.getAngle();
		float orientS2 = targetToBall.multiplyNew(-1).getAngle() - support2ToBall.getAngle();
		
		IVector2 supportPosition = null;
		if (Math.abs(orientS1) < Math.abs(orientS2))
		{
			supportPosition = supportPoint1;
		} else
		{
			supportPosition = supportPoint2;
		}
		
		
		IVector2 behindBall = getWorldFrame().getBall().getPos()
				.addNew(targetToBall.normalizeNew().multiplyNew(safetyDistance));
		
		
		IVector2 advancedSupportPosition = behindBall.addNew(supportPosition.subtractNew(behindBall).multiplyNew(0.5f));
		
		IVector2 finalDestination = getWorldFrame().getBall().getPos()
				.addNew(
						advancedSupportPosition.subtractNew(getWorldFrame().getBall().getPos()).normalizeNew()
								.multiplyNew(safetyDistance));
		
		
		return finalDestination;
	}
	
	
}
