/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Nov 24, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills;

import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.rectangle.Rectangle;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.AIConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.sisyphus.driver.EMovingSpeed;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.ESkillName;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;
import edu.dhbw.mannheim.tigers.sumatra.util.learning.GaussianPolicy;


/**
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 */
public class PolicyTrainerSkill extends PositionSkill
{
	private static final Logger	log			= Logger.getLogger(PolicyTrainerSkill.class.getName());
	
	@Configurable(comment = "standard deviation for generating gaussian samples")
	private static Float[]			posstdev		= { 1f, 1f, 1f };
	
	private final GaussianPolicy	gp;
	private long						lastUpdate	= 0;
	private final Rectangle			field;
	
	
	/**
	 */
	@SuppressWarnings("unused")
	private PolicyTrainerSkill()
	{
		this(GaussianPolicy.createGp("/dev/shm/controller.txt", "/dev/shm/rolloutdata.txt"), AIConfig.getGeometry()
				.getField());
	}
	
	
	/**
	 * @param gp
	 * @param field
	 */
	public PolicyTrainerSkill(final GaussianPolicy gp, final Rectangle field)
	{
		super(ESkillName.POLICY_TRAINER);
		this.gp = gp;
		setCommandType(ECommandType.POS_LEARNED);
		setGpController(gp);
		// limit call of controller to every minDt ms
		// setMinDt(gp.getParams().getDt());
		getMoveCon().setPenaltyAreaAllowedOur(true);
		getMoveCon().setSpeed(EMovingSpeed.UNLIMITED, 3);
		this.field = field;
	}
	
	
	@Override
	protected void doCalcEntryActions(final List<ACommand> cmds)
	{
		genNewDestination();
	}
	
	
	/**
	 * 
	 */
	public void genNewDestination()
	{
		// generate initial state
		Vector2 dest = null;
		for (int j = 0; j < 50; j++)
		{
			dest = new Vector2(getPos());
			for (int i = 0; (i < gp.getParams().getActionDim()) && (i < 2); i++)
			{
				float maxContext = gp.getParams().getMaxContexts().get(i);
				float r = (((float) gp.getRnd().nextGaussian() * (maxContext))) * 1000;
				dest.set(i, dest.get(i) + r);
			}
			if (field.isPointInShape(dest))
			{
				break;
			}
		}
		if (dest == null)
		{
			log.warn("No new destination found...");
		} else
		{
			setDestination(dest);
			
			if (gp.getParams().getActionDim() == 3)
			{
				float orientation = gp.getRnd().nextFloat() * AngleMath.PI_TWO;
				setOrientation(orientation);
			}
		}
	}
	
	
	@Override
	protected void update(final List<ACommand> cmds)
	{
		
		float dt = (SumatraClock.nanoTime() - lastUpdate) * 1e-9f;
		// skip update if we get too many frames
		if (dt >= gp.getDt())
		{
			// reset prob
			float probRef = gp.getRnd().nextFloat();
			if (probRef <= gp.getParams().getResetProb())
			{
				log.debug("Reset " + dt + " " + probRef);
				complete();
			}
			lastUpdate = SumatraClock.nanoTime();
		}
		// float diff = GeoMath.distancePP(getDestination(), getPos());
		// if (diff < 1)
		// {
		// genNewDestination();
		// }
	}
	
	
	@Override
	protected void doCalcExitActions(final List<ACommand> cmds)
	{
	}
}
