/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Mar 9, 2014
 * Author(s): Mark Geiger <Mark.Geiger@dlr.de>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.skills;

import java.util.List;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AiMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.ACommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerKickerKickV2.EKickerMode;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tiger.TigerMotorMoveV2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.commands.tigerv2.TigerSkillPositioningCommand;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.ESkillName;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.EKickDevice;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.skillsystem.devices.TigerDevices;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * Move to a given destination and orientation with PositionController
 * 
 * @author Mark Geiger <Mark.Geiger@dlr.de>
 */
public class PullBackSkill extends ASkill
{
	
	// --------------------------------------------------------------------------
	// --- variables and constants ----------------------------------------------
	// --------------------------------------------------------------------------
	
	@Configurable
	private static long			timeout			= 3000;
	
	private static final long	waitWithPull	= 1200;
	private static final long	waitWithGrab	= 700;
	
	private long					time				= 0;
	private long					waitTimer		= 0;
	
	private EPullState			pullState		= EPullState.GET;
	
	private IVector2				pEndOrient		= null;
	private IVector2				pPullDirection	= null;
	private float					pPullDistance	= 0;
	private float					pSpeed			= 0;
	private IVector2				startPullPos	= null;
	
	private enum EPullState
	{
		GET,
		OPTAIN_BALL,
		PULL
	}
	
	
	// --------------------------------------------------------------------------
	// --- constructors ---------------------------------------------------------
	// --------------------------------------------------------------------------
	
	/**
	 * Do not use this constructor, if you extend from this class
	 * 
	 * @param endOrientation
	 * @param pullDistance
	 * @param pullDirection
	 * @param speed
	 */
	public PullBackSkill(final IVector2 endOrientation, final float pullDistance,
			final IVector2 pullDirection, final float speed)
	{
		super(ESkillName.PULL_BACK);
		pEndOrient = endOrientation;
		pPullDistance = pullDistance;
		pSpeed = speed;
		pPullDirection = pullDirection.normalizeNew();
	}
	
	
	@Override
	protected void doCalcActions(final List<ACommand> cmds)
	{
		if ((SumatraClock.currentTimeMillis() - time) > timeout)
		{
			// complete();
		}
		
		IVector2 dest = null;
		float orient = 0;
		TigerMotorMoveV2 move = null;
		IVector2 localVel = null;
		IVector2 target = null;
		
		switch (pullState)
		{
			case GET:
				dest = calcGetDestination();
				orient = getWorldFrame().getBall().getPos().subtractNew(getPos()).getAngle();
				
				
				if (getWorldFrame().isInverted())
				{
					dest = dest.multiplyNew(-1);
					orient = AngleMath.normalizeAngle(orient + AngleMath.PI);
				}
				
				if (hasBallContact())
				{
					getDevices().dribble(cmds, 18000);
					if (waitTimer < 1)
					{
						waitTimer = SumatraClock.currentTimeMillis();
					}
					else if ((SumatraClock.currentTimeMillis() - waitTimer) > waitWithGrab)
					{
						waitTimer = SumatraClock.currentTimeMillis();
						pullState = EPullState.OPTAIN_BALL;
					}
				} else
				{
					cmds.add(new TigerSkillPositioningCommand(dest, orient));
				}
				break;
			
			case OPTAIN_BALL:
				/*
				 * startSpeed has to be adjusted when pulling sidewise.
				 */
				float startSpeed = -0.1f;
				float endSpeed = pSpeed;
				move = new TigerMotorMoveV2();
				float stime = SumatraClock.currentTimeMillis() - waitTimer;
				
				float finalSpeed = (((endSpeed - startSpeed) / waitWithPull) * stime) + startSpeed;
				
				target = getWorldFrame().getBall().getPos().addNew(pPullDirection.multiplyNew(pPullDistance * 2))
						.subtractNew(getPos());
				localVel = AiMath.convertGlobalBotVector2Local(target, getAngle()).normalizeNew().multiplyNew(finalSpeed);
				
				
				orient = getWorldFrame().getBall().getPos().subtractNew(getPos()).getAngle();
				
				move.setX(localVel.x());
				move.setY(localVel.y());
				
				move.setW(AngleMath.normalizeAngle(orient - getAngle()));
				
				if ((SumatraClock.currentTimeMillis() - waitTimer) > waitWithPull)
				{
					pullState = EPullState.PULL;
					waitTimer = SumatraClock.currentTimeMillis();
				}
				getDevices().dribble(cmds, 18000);
				cmds.add(move);
				
				break;
			
			case PULL:
				
				target = startPullPos.addNew(pPullDirection.multiplyNew(pPullDistance * 2)).subtractNew(getPos());
				localVel = AiMath.convertGlobalBotVector2Local(target, getAngle()).normalizeNew().multiplyNew(pSpeed);
				if (GeoMath.distancePP(startPullPos, getPos()) > pPullDistance)
				{
					getDevices().kickGeneral(cmds, EKickerMode.FORCE, EKickDevice.STRAIGHT,
							TigerDevices.getStraightKickMaxDuration(), 0);
					complete();
				}
				
				move = new TigerMotorMoveV2();
				orient = pEndOrient.getAngle();
				
				move.setX(localVel.x());
				move.setY(localVel.y());
				
				move.setW(AngleMath.normalizeAngle(orient - getAngle()));
				
				cmds.add(move);
				break;
			
			default:
				break;
		}
	}
	
	
	@Override
	protected void doCalcEntryActions(final List<ACommand> cmds)
	{
		time = SumatraClock.currentTimeMillis();
		startPullPos = getWorldFrame().getBall().getPos();
	}
	
	
	private IVector2 calcGetDestination()
	{
		IVector2 destination = null;
		IVector2 ballPos = getWorldFrame().getBall().getPos();
		IVector2 rvBallBot = getPos().subtractNew(ballPos).normalizeNew();
		float distanceControlledSpeed = GeoMath.distancePP(ballPos, getPos()) / 2;
		destination = ballPos.addNew(rvBallBot.multiplyNew(distanceControlledSpeed));
		return destination;
	}
}
