/*
 * *********************************************************
 * Copyright (c) 2009 - 2011, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: May 15, 2011
 * Author(s): Birgit
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.kalman.oextkal;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AiMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.AngleMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.cam.CamBall;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.cam.CamRobot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.AIConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.Geometry;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.kalman.oextkal.flyingBalls.Altigraph;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;
import edu.dhbw.mannheim.tigers.sumatra.util.config.Configurable;


/**
 * This Class is to correct the ssl-ballPosition from bottom to fly
 * 
 * @author Birgit
 */
public class BallCorrector
{
	@SuppressWarnings("unused")
	private static final Logger	log						= Logger.getLogger(BallCorrector.class.getName());
	
	@Configurable(comment = "correct flying balls")
	private static boolean			correctFlyingBalls	= false;
	private static final float		MAX_DIST_BALL			= 50;
	private static final float		MAX_ORIENTATION_DIFF	= 0.1f;
	
	private Altigraph					m_a;
	
	private CamBall					lastSeenBall			= CamBall.defaultInstance();
	
	
	/**
	 */
	public BallCorrector()
	{
		m_a = new Altigraph();
	}
	
	
	/**
	 * correct the ball frame, if ball is flying
	 * 
	 * @param balls
	 * @param yellowBots
	 * @param blueBots
	 * @return
	 */
	public CamBall correctBall(final List<CamBall> balls, final List<CamRobot> yellowBots, final List<CamRobot> blueBots)
	{
		CamBall ballToUse = findCurrentBall(balls);
		
		if (correctFlyingBalls)
		{
			// get Data of Ball and Bots
			final List<CamRobot> bots = new ArrayList<>(blueBots);
			bots.addAll(yellowBots);
			
			// if kick possible, add new fly
			CamRobot possShooter = findPossibleShooter(ballToUse, bots);
			if (possShooter != null)
			{
				// append new fly
				// log.trace("kicker zone identified: " + possShooter.getPos() + " " + possShooter.getOrientation());
				m_a.addKickerZoneIdentified(possShooter.getPos().x(), possShooter.getPos().y(),
						possShooter.getOrientation());
			}
			
			// append the ball to old and new flys
			m_a.addCamFrame(ballToUse.getPos().x(), ballToUse.getPos().y(), ballToUse.getCameraId());
			
			// if ball is flying
			if (m_a.isBallFlying())
			{
				// log.trace("Ball is flying: " + m_a.getCorrectedFrame().z());
				CamBall newBall = new CamBall(ballToUse.getConfidence(), ballToUse.getArea(), (float) m_a
						.getCorrectedFrame()
						.x(),
						(float) m_a.getCorrectedFrame().y(), (float) m_a.getCorrectedFrame().z(), ballToUse.getPixelX(),
						ballToUse.getPixelY(), ballToUse.getTimestamp(), ballToUse.getCameraId());
				return newBall;
			}
		}
		return ballToUse;
	}
	
	
	private CamBall findCurrentBall(final List<CamBall> balls)
	{
		if (balls.isEmpty())
		{
			if ((System.nanoTime() - lastSeenBall.getTimestamp()) > 3e8)
			{
				lastSeenBall = new CamBall(0, 0, lastSeenBall.getPos().x(), lastSeenBall.getPos().y(), lastSeenBall
						.getPos().z(),
						lastSeenBall.getPixelX(), lastSeenBall.getPixelY(), System.nanoTime(),
						lastSeenBall.getCameraId());
			}
			return lastSeenBall;
		}
		
		CamBall ballToUse = balls.get(0);
		float shortestDifference = difference(ballToUse, lastSeenBall);
		for (CamBall ball : balls)
		{
			if (difference(ball, lastSeenBall) < shortestDifference)
			{
				ballToUse = ball;
				shortestDifference = difference(ball, lastSeenBall);
			}
		}
		// 1mm -> 0.1ms
		// 10mm -> 1ms -> 10m/s
		// if the ball is moving with less than 10m/s, the new ball will always be used. (following if will be false)
		if (difference(lastSeenBall, ballToUse) > (TimeUnit.NANOSECONDS.toMillis(SumatraClock.nanoTime()
				- lastSeenBall.getTimestamp()) * 10))
		{
			return lastSeenBall;
		}
		lastSeenBall = new CamBall(ballToUse);
		return ballToUse;
	}
	
	
	private float difference(final CamBall ball1, final CamBall ball2)
	{
		if (ball1.getTimestamp() > (ball2.getTimestamp() + 1e8))
		{
			return 0;
		}
		return ball1.getPos().subtractNew(ball2.getPos()).getLength2()
				// add a penalty to balls that are not on same camera
				+ (ball1.getCameraId() == ball2.getCameraId() ? 0 : 500);
	}
	
	
	/*
	 * find out the potential kicker-bot
	 */
	private CamRobot findPossibleShooter(final CamBall ball, final List<CamRobot> bots)
	{
		for (final CamRobot bot : bots)
		{
			IVector2 kickerPos = AiMath
					.getBotKickerPos(bot.getPos(), bot.getOrientation(), Geometry.getCenter2DribblerDistDefault());
			if (GeoMath.distancePP(kickerPos, ball.getPos().getXYVector()) > (MAX_DIST_BALL + AIConfig.getGeometry()
					.getBotRadius()))
			{
				continue;
			}
			IVector2 bot2Ball = ball.getPos().getXYVector().subtract(bot.getPos());
			if (Math.abs(AngleMath.difference(bot2Ball.getAngle(), bot.getOrientation())) > MAX_ORIENTATION_DIFF)
			{
				continue;
			}
			
			return bot;
		}
		return null;
	}
	
	
	/**
	 * @param pos
	 */
	public void setLatestBallPos(final IVector2 pos)
	{
		lastSeenBall = new CamBall(1, 0, pos.x(), pos.y(), 0, 0, 0, System.nanoTime(), 0);
	}
}
