/*
 * *********************************************************
 * Copyright (c) 2009 - 2011, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: May 17, 2011
 * Author(s): Birgit
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.kalman.oextkal.flyingBalls;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.cam.Coord;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.AIConfig;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.ai.config.Geometry;


/**
 * One possible Fly
 * 
 * @author Birgit
 */
public class Fly
{
	@SuppressWarnings("unused")
	private static final Logger	log								= Logger.getLogger(Fly.class.getName());
	
	private static final int		MAX_NUMBER_BALLS_GO_BACK	= 4;
	/** mm */
	private static final double	MIN_FLY_HEIGHT					= 50;
	/** mm */
	private static final double	MIN_FLY_LENGTH					= 400;
	/** */
	private static final double	RUN_BOT2BALL_MAX_ANGLE		= Math.PI / 6;
	/** mm */
	private static final double	BALL2BALL_MIN_DISTANCE		= 10;
	
	private List<FlyingBall>		balls								= new ArrayList<FlyingBall>();
	private final Coord				kickPos;
	private final LinFunc			roboViewFunction;
	private RegParab					flyParabel						= null;
	private double						lastDistance					= 0;
	private int							backCount						= 0;
	
	
	/**
	 * create a new fly
	 * 
	 * @param aRoboPos
	 * @param aViewAngle
	 */
	public Fly(final Coord aRoboPos, final double aViewAngle)
	{
		// function through roboter and angle
		roboViewFunction = new LinFunc(aRoboPos, aViewAngle);
		// set the kickingPosition
		kickPos = roboViewFunction.goDistanceFromPoint(aRoboPos, Geometry.getCenter2DribblerDistDefault());
	}
	
	
	/**
	 * add a new ball and control, weather it has an orientation between given borders
	 * 
	 * @param aBallPos
	 * @param camID
	 * @return
	 */
	public boolean addNewBall(final Coord aBallPos, final int camID)
	{
		// Create new ballInstance
		final FlyingBall ball = new FlyingBall(aBallPos);
		
		// set the line through the ball and the cam
		LinFunc ballOrientation;
		
		float prinX = AIConfig.getGeometry().getCameraPrincipalPointX()[camID];
		float prinY = AIConfig.getGeometry().getCameraPrincipalPointY()[camID];
		ballOrientation = new LinFunc(new Coord(prinX, prinY), aBallPos);
		
		// look for cut with robot-line, which is flying position
		final Coord flyingPosition = LinFunc.getCutCoords(roboViewFunction, ballOrientation);
		ball.setFlyPositionAndCalculateFlyingHeight(flyingPosition, camID);
		ball.calculateDistanceToStart(kickPos);
		
		return doesBallFit(ball);
	}
	
	
	/*
	 * check, weather the vector between current and before ball is between borders,
	 * given in definitions
	 */
	private boolean doesBallFit(final FlyingBall flyingBall)
	{
		final Coord ball = flyingBall.getBottomPosition();
		final double distStartBallX = ball.x() - kickPos.x();
		final double distStartBallY = ball.y() - kickPos.y();
		final Coord distanceStartBall = new Coord(distStartBallX, distStartBallY);
		final double newDis = distanceStartBall.getLength();
		final double angle = roboViewFunction.getAngleToVector(distanceStartBall);
		
		
		// check for longest distance
		if (flyParabel != null)
		{
			final double maxDis = -flyParabel.getD() * 2;
			if (balls.size() == 1)
			{
				lastDistance = newDis;
			}
			
			if (newDis < lastDistance)
			{
				backCount++;
			}
			
			if ((newDis > maxDis) || (backCount >= MAX_NUMBER_BALLS_GO_BACK))
			{
				return false;
			}
			
			lastDistance = newDis;
		}
		
		
		// angle fits
		if (angle < RUN_BOT2BALL_MAX_ANGLE)
		{
			// -- check for long enough distance was passed
			
			// calculate lastPos
			Coord lastBall;
			// if we have the first ball, get the position from start
			if (balls.isEmpty())
			{
				lastBall = kickPos;
			}
			// if we have many balls, get the previous ball-position
			else
			{
				lastBall = balls.get(balls.size() - 1).getBottomPosition();
			}
			final double distBallBallX = ball.x() - lastBall.x();
			final double distBallBallY = ball.y() - lastBall.y();
			
			final Coord ball2ballDistance = new Coord(distBallBallX, distBallBallY);
			
			if (ball2ballDistance.getLength() <= BALL2BALL_MIN_DISTANCE)
			{
				// it probably fits, but this one ball is not useful for us...
				return true;
			}
		}
		// angle fits not
		else
		{
			return false;
		}
		balls.add(flyingBall);
		return true;
	}
	
	
	/**
	 * calculate the regressionParabel
	 */
	public void calculateFly()
	{
		
		// if no parabel exists
		if (flyParabel == null)
		{
			final int number = balls.size();
			final Coord[] flyingBalls = new Coord[number];
			
			// getAllData
			for (int i = 0; i < number; i++)
			{
				final FlyingBall ball = balls.get(i);
				flyingBalls[i] = new Coord(ball.getDistance(), ball.getFlyingHeight());
			}
			flyParabel = new RegParab(flyingBalls);
		} else
		{
			final int last = balls.size() - 1;
			flyParabel.appendData(new Coord(balls.get(last).getDistance(), balls.get(last).getFlyingHeight()));
		}
	}
	
	
	/**
	 * return, weather minimum high at least is reached
	 * 
	 * @return
	 */
	public boolean isAtLeastMinHighReached()
	{
		if ((flyParabel.getE() > MIN_FLY_HEIGHT) && ((-flyParabel.getD()) > MIN_FLY_LENGTH))
		{
			return true;
		}
		return false;
	}
	
	
	// --------------------------------------------------------------------------
	// --- getter/setter --------------------------------------------------------
	// --------------------------------------------------------------------------
	/**
	 * @return
	 */
	public int size()
	{
		return balls.size();
	}
	
	
	/**
	 * @return
	 */
	public Coord getStartPos()
	{
		return kickPos;
	}
	
	
	/**
	 * @return
	 */
	public RegParab getParabel()
	{
		return flyParabel;
	}
	
	
	/**
	 * @return
	 */
	public Coord getCurrentBallPosition()
	{
		return balls.get(balls.size() - 1).getFlyingPosition();
	}
	
	
	/**
	 * @return
	 */
	public double getCurrentBallHeight()
	{
		return balls.get(balls.size() - 1).getFlyingHeight();
	}
	
	
	@Override
	public String toString()
	{
		final StringBuilder str = new StringBuilder();
		
		str.append("Fly: ");
		for (int i = 0; i < balls.size(); i++)
		{
			str.append("Ball " + i + ":\n");
			str.append(balls.get(i).toString());
		}
		
		return str.toString();
	}
	
	
	/**
	 * @return the balls
	 */
	public List<FlyingBall> getBalls()
	{
		return balls;
	}
	
	
}
