/*
 * *********************************************************
 * Copyright (c) 2009 - 2014, DHBW Mannheim - Tigers Mannheim
 * Project: TIGERS - Sumatra
 * Date: Mar 11, 2014
 * Author(s): Nicolai Ommer <nicolai.ommer@gmail.com>
 * *********************************************************
 */
package edu.dhbw.mannheim.tigers.sumatra.util.csvexporter;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import org.apache.log4j.Logger;

import edu.dhbw.mannheim.tigers.moduli.exceptions.ModuleNotFoundException;
import edu.dhbw.mannheim.tigers.sumatra.model.SumatraModel;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.SimpleWorldFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.data.frames.WorldFrameWrapper;
import edu.dhbw.mannheim.tigers.sumatra.model.data.math.GeoMath;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.ai.ETeamColor;
import edu.dhbw.mannheim.tigers.sumatra.model.data.modules.cam.CamBall;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.IVector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector2;
import edu.dhbw.mannheim.tigers.sumatra.model.data.shapes.vector.Vector3;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.TrackedTigerBot;
import edu.dhbw.mannheim.tigers.sumatra.model.data.trackedobjects.ids.BotID;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.botmanager.bots.DummyBot;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.IMergedCamFrameObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.impls.worldpredictor.MergedCamDetectionFrame;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.observer.IWorldPredictorObserver;
import edu.dhbw.mannheim.tigers.sumatra.model.modules.types.AWorldPredictor;
import edu.dhbw.mannheim.tigers.sumatra.util.clock.SumatraClock;


/**
 * Watch chip kicks and gather data to learn for future
 * 
 * @author Nicolai Ommer <nicolai.ommer@gmail.com>
 * @author Mark Geiger <Mark.Geiger@dlr.de>
 */
public class BallWatcher implements IWorldPredictorObserver, Runnable, IMergedCamFrameObserver
{
	private static final Logger					log						= Logger
																								.getLogger(BallWatcher.class.getName());
	private static final long						TIMEOUT_NS				= (long) 10e9;
	private final long								startTime				= SumatraClock.nanoTime();
	private IVector2									currentVel				= Vector2.ZERO_VECTOR;
	private IVector3									currentPos				= Vector3.ZERO_VECTOR;
	private SimpleWorldFrame						currentFrame			= null;
	private CamBall									lastBall					= null;
	private IVector2									initBallPos				= null;
	private List<List<Object>>						data						= new LinkedList<>();
	private final String								fileName;
	private long										time2Stop				= 0;
	private boolean									processing				= false;
	private final List<IBallWatcherObserver>	observers				= new CopyOnWriteArrayList<IBallWatcherObserver>();
	private int											numFramesBallStopped	= 0;
	
	
	/**
	 * @param fileName
	 */
	public BallWatcher(final String fileName)
	{
		String moduli = SumatraModel.getInstance().getCurrentModuliConfig().split("\\.")[0];
		this.fileName = moduli + "/" + fileName;
	}
	
	
	/**
	 * @param observer
	 */
	public final void addObserver(final IBallWatcherObserver observer)
	{
		observers.add(observer);
	}
	
	
	/**
	 * @param observer
	 */
	public final void removeObserver(final IBallWatcherObserver observer)
	{
		observers.remove(observer);
	}
	
	
	private void notifyExport(final CSVExporter event)
	{
		for (IBallWatcherObserver observer : observers)
		{
			observer.beforeExport(event);
		}
	}
	
	
	private void notifyCustomData(final List<Object> event)
	{
		for (IBallWatcherObserver observer : observers)
		{
			observer.onAddCustomData(event);
		}
	}
	
	
	private void notifyPostProcessing(final String filename)
	{
		for (IBallWatcherObserver observer : observers)
		{
			observer.postProcessing(filename);
		}
	}
	
	
	private void exportData()
	{
		stop();
		if (!processing)
		{
			processing = true;
			new Thread(this, "BallWatcher").start();
		}
	}
	
	
	private void stop()
	{
		try
		{
			AWorldPredictor wp = (AWorldPredictor) SumatraModel.getInstance().getModule(AWorldPredictor.MODULE_ID);
			wp.removeObserver(this);
			wp.removeMergedFrameObserver(this);
		} catch (ModuleNotFoundException err)
		{
			log.error("WP module not found.", err);
		}
	}
	
	
	/**
	 * 
	 */
	public final void start()
	{
		try
		{
			AWorldPredictor wp = (AWorldPredictor) SumatraModel.getInstance().getModule(AWorldPredictor.MODULE_ID);
			wp.addObserver(this);
			wp.addMergedFrameObserver(this);
		} catch (ModuleNotFoundException err)
		{
			log.error("WP module not found.", err);
		}
	}
	
	
	protected boolean checkIsFailed(final CamBall ball)
	{
		return false;
	}
	
	
	protected boolean checkIsDone(final CamBall ball)
	{
		if (lastBall.getPos().getXYVector().equals(ball.getPos().getXYVector(), 5)
				&& (!initBallPos.equals(ball.getPos().getXYVector(), 50))
				&& (currentVel.getLength2() < 0.01f))
		{
			numFramesBallStopped++;
			if (numFramesBallStopped > 10)
			{
				log.debug("ball stopped, data size: " + data.size());
				return true;
			}
		} else
		{
			numFramesBallStopped = 0;
		}
		if ((time2Stop != 0) && ((SumatraClock.nanoTime() - time2Stop) > 0))
		{
			log.debug("requested timeout reached. Done.");
			return true;
		}
		if ((SumatraClock.nanoTime() - startTime) > TIMEOUT_NS)
		{
			log.debug("Ball watcher timed out");
			return true;
		}
		return false;
	}
	
	
	@Override
	public void onNewCameraFrame(final MergedCamDetectionFrame frame)
	{
		// do not collect more data if we are already processing the data!
		if (processing)
		{
			return;
		}
		if (currentFrame == null)
		{
			return;
		}
		
		CamBall curBall = frame.getBall();
		if (lastBall == null)
		{
			lastBall = curBall;
		}
		
		if (lastBall.getTimestamp() == curBall.getTimestamp())
		{
			return;
		}
		
		if (initBallPos == null)
		{
			initBallPos = curBall.getPos().getXYVector();
		}
		TrackedTigerBot nearestBot = getBotNearestToBall(currentFrame, curBall);
		IVector2 acc = currentFrame.getBall().getAcc();
		
		List<Object> dataSet = new ArrayList<Object>();
		dataSet.add(System.nanoTime());
		dataSet.add(curBall.getTimestamp());
		dataSet.add(curBall.getCameraId());
		dataSet.add(curBall.getPos().x());
		dataSet.add(curBall.getPos().y());
		dataSet.add(currentVel.x());
		dataSet.add(currentVel.y());
		dataSet.add(currentVel.getLength2());
		dataSet.add(currentPos.x());
		dataSet.add(currentPos.y());
		dataSet.add(currentPos.z());
		dataSet.add(acc.x());
		dataSet.add(acc.y());
		dataSet.add(nearestBot.getPos().x());
		dataSet.add(nearestBot.getPos().y());
		dataSet.add(nearestBot.getAngle());
		notifyCustomData(dataSet);
		data.add(dataSet);
		
		lastBall = curBall;
		
		if (checkIsFailed(curBall))
		{
			stop();
			data.clear();
			return;
		}
		
		if (checkIsDone(curBall))
		{
			exportData();
		}
	}
	
	
	private TrackedTigerBot getBotNearestToBall(final SimpleWorldFrame frame, final CamBall curBall)
	{
		IVector2 ballPos = curBall.getPos().getXYVector();
		float minDist = Float.MAX_VALUE;
		TrackedTigerBot nearest = TrackedTigerBot.defaultBot(BotID.createBotId(0, ETeamColor.BLUE), new DummyBot());
		for (TrackedTigerBot bot : frame.getBots().values())
		{
			float dist = GeoMath.distancePP(ballPos, bot.getPos());
			if (dist < minDist)
			{
				nearest = bot;
				minDist = dist;
			}
		}
		return nearest;
	}
	
	
	/**
	 * Stop after given delay
	 * 
	 * @param milliseconds
	 */
	public final void stopDelayed(final long milliseconds)
	{
		time2Stop = SumatraClock.nanoTime() + (milliseconds * (long) 1e6);
	}
	
	
	/**
	 * 
	 */
	public final void stopExport()
	{
		stopDelayed(0);
	}
	
	
	@Override
	public void onNewWorldFrame(final WorldFrameWrapper wfWrapper)
	{
		currentVel = wfWrapper.getSimpleWorldFrame().getBall().getVel();
		currentPos = wfWrapper.getSimpleWorldFrame().getBall().getPos3();
		currentFrame = wfWrapper.getSimpleWorldFrame();
	}
	
	
	@Override
	public final void run()
	{
		CSVExporter exporter = new CSVExporter(fileName, false);
		
		for (List<Object> ds : data)
		{
			exporter.addValues(ds);
		}
		
		notifyExport(exporter);
		
		exporter.close();
		data.clear();
		
		notifyPostProcessing(exporter.getAbsoluteFileName());
	}
	
	
	/**
	 * @return the initBallPos
	 */
	public final IVector2 getInitBallPos()
	{
		return initBallPos;
	}
	
	
	/**
	 * @return the lastBall
	 */
	public final CamBall getLastBall()
	{
		return lastBall;
	}
	
	
	/**
	 * @return the fileName
	 */
	public final String getFileName()
	{
		return fileName;
	}
}
